#!/usr/bin/env python

import os
import time
import gc
import yaml

from typing import List, Union

import torch
import uvicorn

from fastapi import FastAPI, Request, Body
from fastapi.responses import HTMLResponse
from pydantic import BaseModel, Field, model_validator
from jinja2.exceptions import TemplateError
from jinja2.sandbox import ImmutableSandboxedEnvironment

import eole
from eole.inference_engine import InferenceEnginePY
from eole.config.run import PredictConfig
from eole.config.inference import DecodingConfig
from eole.bin import register_bin, BaseBin
from eole.utils.logging import logger
from eole.constants import DefaultTokens

STATUS_OK = "ok"
STATUS_ERROR = "error"


class TextRequest(DecodingConfig):
    """
    Standard text "completion" request
    (as well as encoder/decoder models e.g. translation).
    """

    model: int | str = Field(description="Model identifier from server configuration.")
    inputs: Union[str, List[str]] = Field(
        description="List of inputs to run inference on. "
        "A single string will be automatically cast to a single item list."
    )

    class Config:
        json_schema_extra = {
            "example": {
                "model": "llama3-8b-instruct",
                "inputs": "<|begin_of_text|><|start_header_id|>system<|end_header_id|>You are a funny guy.<|eot_id|><|start_header_id|>user<|end_header_id|>Tell me a joke :)<|eot_id|><|start_header_id|>assistant<|end_header_id|>",  # noqa: E501
            }
        }


class TextResponse(BaseModel):
    """
    Response of TextRequest.
    """

    predictions: List[List[str]] = Field(description="List of prediction(s) for each input(s).")
    scores: List[List[float]] = Field(description="Pred scores from the model for each prediction.")

    class Config:
        json_schema_extra = {
            "example": {
                "predictions": [
                    [
                        "\n\nHere's one:\n\nWhy couldn't the bicycle stand up by itself?\n\n(wait for it...)\n\nBecause it was two-tired!\n\nHope that made you laugh!"  # noqa: E501
                    ]
                ],
                "scores": [[-0.040771484375]],
            }
        }

    @model_validator(mode="after")
    def _validate_response(self):
        """
        Automatically apply some formatting to the provided text response.
        This logic might be moved elsewhere at some point.
        """
        self.predictions = [[pred.replace(DefaultTokens.SEP, "\n") for pred in preds] for preds in self.predictions]
        return self


class ChatRequest(DecodingConfig):
    """
    Request format for chat-based interactions.
    """

    model: int | str = Field(description="Model identifier from server configuration.")
    messages: List[dict] = Field(description="List of message dictionaries with 'role' and 'content' keys.")

    class Config:
        json_schema_extra = {
            "example": {
                "model": "llama3-8b-instruct",
                "messages": [
                    {"role": "system", "content": "You are a funny guy."},
                    {"role": "user", "content": "Tell me a joke :)"},
                ],
            }
        }


# TODO: specific response model for chat mode?
# class ChatResponse(BaseModel):
#     choices: List[dict]


class Server(object):
    """
    Main server class to manage configuration, models and corresponding constraints.
    """

    def __init__(self):
        self.start_time = time.time()
        self.models = {}
        self.models_root = None

    def start(self, server_config_path):
        """
        Initialize the server with the given configuration.
        """
        with open(server_config_path) as f:
            server_config = yaml.safe_load(f)
        self.models_root = server_config["models_root"]
        for model in server_config["models"]:
            # instantiate models
            model_id = model["id"]
            model_path = model["path"]
            self.models[model_id] = Model(
                model_id=model_id,
                model_path=model_path,
                models_root=self.models_root,
                model_type=model.get("model_type", "default"),
                pre_config=model.get("config", {}),
            )
            if model.get("preload", False):
                self.models[model_id].load()

    def available_models(self):
        """
        Return a list of available models.
        """
        models = []
        for model_id, model in self.models.items():
            models.append({"id": model_id})
        return models

    def maybe_load_model(self, model_id_to_load):
        """
        Very naive method to ensure a single model is loaded for now.
        """
        for model_id, model in self.models.items():
            if model_id != model_id_to_load:
                model.unload()


class Model(object):
    """
    Represents a single model in the server.
    """

    def __init__(
        self,
        model_id=None,
        model_path=None,
        preload=False,
        models_root=None,
        model_type=False,
        pre_config={},
    ):
        self.loaded = False
        self.engine = None
        self.model_id = model_id
        self.preload = preload
        self.models_root = models_root
        self.model_path = model_path
        self.local_path = None
        self.model_type = model_type
        self.pre_config = pre_config

    def get_config(self):
        """
        Instanciate the configuration for the model.
        """
        # transforms and inference settings are retrieved from the model config for now
        self.config = PredictConfig(
            src="dummy",
            model_path=self.local_path,
            # TODO improve this
            gpu_ranks=[0],
            world_size=1,
            **self.pre_config,
        )

    def maybe_retrieve_model(self):
        """
        Download the model if it's not available locally.
        """
        from huggingface_hub import HfApi, snapshot_download

        hf_api = HfApi()
        try:
            hf_api.model_info(self.model_path)
        except Exception:
            self.local_path = os.path.expandvars(self.model_path)
        else:
            self.local_path = os.path.expandvars(os.path.join(self.models_root, self.model_path))
            logger.info(f"Downloading {self.model_path} from huggingface, " f"to local directory {self.local_path}")
            snapshot_download(repo_id=self.model_path, local_dir=self.local_path)

    def load(self):
        """
        Create the inference engine.
        """
        self.maybe_retrieve_model()
        self.get_config()
        self.engine = InferenceEnginePY(self.config)
        self.loaded = True
        logger.info(f"Loaded model {self.model_id} from: {self.model_path}")

    def unload(self):
        """
        Not super clean, we might want to do better some day...
        """
        del self.engine
        gc.collect()
        torch.cuda.empty_cache()
        self.engine = None
        self.loaded = False
        logger.info(f"Unloaded model {self.model_id}")

    def apply_chat_template(self, inputs):
        """
        Render the model input based on the model chat template
        and the request inputs.
        """

        def raise_exception(message):
            raise TemplateError(message)

        jinja_env = ImmutableSandboxedEnvironment(trim_blocks=True, lstrip_blocks=True)
        jinja_env.globals["raise_exception"] = raise_exception
        template = jinja_env.from_string(self.config.chat_template)
        rendered_output = template.render(
            **{
                "messages": inputs,
                "bos_token": "",  # handled in numericalize
                "add_generation_prompt": True,
            }
        )
        return rendered_output

    def infer(self, inputs, settings={}, is_chat=False):
        """
        Run inference on the given inputs.
        """
        if isinstance(inputs, str):
            inputs = [inputs]
        if not (self.loaded):
            self.load()
        if is_chat:
            assert hasattr(self.config, "chat_template"), "Chat requests can't be performed without a chat_template."
            inputs = [self.apply_chat_template(inputs)]
        scores, _, preds = self.engine.infer_list(inputs, settings=settings)
        return scores, preds


def create_app(config_file):
    """
    Create and configure the FastAPI application.
    """
    app = FastAPI(
        title="Eole Inference Server",
        version=eole.__version__,
        summary="A simple inference server to expose various models.",
        description="",  # TODO
    )

    server = Server()
    server.start(config_file)

    @app.get("/")
    def root(request: Request):
        """
        Root endpoint returning HTML content to help users find the docs.
        """
        html_content = f"""
        <html>
            <head>
                <title>Eole Server</title>
            </head>
            <body>
                <h1>Eole Server</h1>
                <p>Probably not what you're looking for.</p>
                <p>API docs --> <a href="{request.url}docs">{request.url}docs</a>.</p>
            </body>
        </html>
        """
        return HTMLResponse(content=html_content, status_code=200)

    @app.get("/models")
    def models():
        """
        Return available models currently exposed.
        """
        models = server.available_models()
        out = {"models": models}
        return out

    @app.post("/unload_model")
    def unload_model(model_id):
        """
        Unload a specific model.
        """
        server.models[model_id].unload()

    @app.get("/health")
    def health():
        """
        Health check endpoint.
        """
        out = {}
        out["status"] = STATUS_OK
        return out

    @app.post("/infer", response_model=TextResponse)
    def infer(
        request: Union[TextRequest, ChatRequest] = Body(
            openapi_examples={
                "text_request": {
                    "summary": "Text Request Example",
                    "description": "A sample text request",
                    "value": TextRequest.Config.json_schema_extra["example"],
                },
                "chat_request": {
                    "summary": "Chat Request Example",
                    "description": "A sample chat request",
                    "value": ChatRequest.Config.json_schema_extra["example"],
                },
            },
        ),
    ):
        """
        Run inference on the given input.
        """
        if isinstance(request, TextRequest):
            inputs = request.inputs if isinstance(request.inputs, list) else [request.inputs]
        else:  # ChatRequest
            # no batch support right now
            inputs = request.messages
        model_id = request.model
        # automatically grab anything that is not model/inputs
        # (we could probably rely on pydantic model once properly implemented)
        non_settings_keys = ["inputs", "messages", "model"]
        settings = {k: v for k, v in request.model_dump().items() if k not in non_settings_keys}
        # TODO: move this in some `infer` method in the `Server` class?
        server.maybe_load_model(model_id)
        scores, preds = server.models[model_id].infer(
            inputs,
            settings=settings,
            is_chat=isinstance(request, ChatRequest),
        )
        # returned scores are tensors which we need to cast (not anymore?)
        # scores = [[score.item() for score in score_list] for score_list in scores]
        response = {"predictions": preds, "scores": scores}
        return response

    # @app.post("/openai/chat/completions")
    # def openai_chat(request: ChatRequest):
    #     """
    #     Simulate an OpenAI Request.
    #     The idea is to make this a viable alternative as a drop-in
    #     replacement for OpenAI or other LLM stacks.
    #     """
    #     pass

    return app


@register_bin(name="serve")
class Serve(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--config",
            "-config",
            "-c",
            default="./server_conf.yaml",
            help="Path of server YAML config file.",
        )
        parser.add_argument("--host", type=str, default="0.0.0.0")
        parser.add_argument("--port", type=int, default="5000")

    @classmethod
    def run(cls, args):
        app = create_app(args.config)
        uvicorn.run(app=app, host=args.host, port=args.port, log_level="info")



""" Optimizers class """

import torch
from torch.amp import GradScaler
from torch.nn.utils import clip_grad_norm_
import operator
import functools
from copy import copy
from math import sqrt, cos, pi
import os

try:
    import optimi

    optimi_available = True
except ImportError:
    optimi_available = False
    pass


def build_torch_optimizer(model, config):
    """Builds the PyTorch optimizer.

    We use the default parameters for Adam that are suggested by
    the original paper https://arxiv.org/pdf/1412.6980.pdf
    These values are also used by other established implementations,
    e.g. https://www.tensorflow.org/api_docs/python/tf/train/AdamOptimizer
    https://keras.io/optimizers/
    Recently there are slightly different values used in the paper
    "Attention is all you need"
    https://arxiv.org/pdf/1706.03762.pdf, particularly the value beta2=0.98
    was used there however, beta2=0.999 is still arguably the more
    established value, so we use that here as well

    Args:
      model: The model to optimize.
      opt. The dictionary of options.

    Returns:
      A ``torch.optim.Optimizer`` instance.
    """
    params = [p for p in model.parameters() if p.requires_grad]
    betas = [config.adam_beta1, config.adam_beta2]

    if config.use_amp or not optimi_available:
        optim = torch.optim
    else:
        optim = optimi
    # optimi supports only sgd / adam / adamw for us
    # hence we use directly torch.optim for others
    if config.optim == "sgd":
        optimizer = optim.SGD(params, lr=config.learning_rate, weight_decay=config.weight_decay)
    elif config.optim == "adagrad":
        optimizer = torch.optim.Adagrad(
            params,
            lr=config.learning_rate,
            initial_accumulator_value=config.adagrad_accumulator_init,
            weight_decay=config.weight_decay,
        )
    elif config.optim == "adadelta":
        optimizer = torch.optim.Adadelta(params, lr=config.learning_rate, weight_decay=config.weight_decay)
    elif config.optim == "adafactor":
        optimizer = AdaFactor(
            params,
            non_constant_decay=True,
            enable_factorization=True,
            weight_decay=config.weight_decay,
        )
    elif config.optim == "adam":
        optimizer = optim.Adam(
            params,
            lr=config.learning_rate,
            betas=betas,
            eps=config.adam_eps,
            weight_decay=config.weight_decay,
        )
    elif config.optim == "adamw":
        optimizer = optim.AdamW(
            params,
            lr=config.learning_rate,
            betas=betas,
            eps=config.adam_eps,
            weight_decay=config.weight_decay,
            foreach=False,
        )
    elif config.optim == "sparseadam":
        dense = []
        sparse = []
        for name, param in model.named_parameters():
            if not param.requires_grad:
                continue
            # TODO: Find a better way to check for sparse gradients.
            if "embed" in name:
                sparse.append(param)
            else:
                dense.append(param)
        optimizer = MultipleOptimizer(
            [
                optim.Adam(dense, lr=config.learning_rate, betas=betas, eps=config.adam_eps),
                torch.optim.SparseAdam(sparse, lr=config.learning_rate, betas=betas, eps=config.adam_eps),
            ]
        )
    elif config.optim in ["adamw8bit", "pagedadamw8bit", "pagedadamw32bit"]:
        try:
            os.environ["BITSANDBYTES_NOWELCOME"] = "1"
            import bitsandbytes as bnb
        except ImportError:
            raise ImportError("Install bitsandbytes to use bnb optimizers")
        if config.optim == "adamw8bit":
            optimizer = bnb.optim.AdamW8bit(
                params,
                lr=config.learning_rate,
                betas=betas,
                eps=config.adam_eps,
                weight_decay=config.weight_decay,
                amsgrad=False,
                optim_bits=8,
                args=None,
                min_8bit_size=1024,
                percentile_clipping=100,
                block_wise=True,
                is_paged=False,
            )
        elif config.optim == "pagedadamw8bit":
            optimizer = bnb.optim.PagedAdamW8bit(
                params,
                lr=config.learning_rate,
                betas=betas,
                eps=config.adam_eps,
                weight_decay=config.weight_decay,
                amsgrad=False,
                optim_bits=8,
                args=None,
                min_8bit_size=4096,
                percentile_clipping=100,
                block_wise=True,
            )
        elif config.optim == "pagedadamw32bit":
            optimizer = bnb.optim.PagedAdamW32bit(
                params,
                lr=config.learning_rate,
                betas=betas,
                eps=config.adam_eps,
                weight_decay=config.weight_decay,
                amsgrad=False,
                optim_bits=32,
                args=None,
                min_8bit_size=4096,
                percentile_clipping=100,
                block_wise=True,
            )
        else:
            raise ValueError("Invalid optimizer type: " + config.optim)
    else:
        raise ValueError("Invalid optimizer type: " + config.optim)

    return optimizer


def make_learning_rate_decay_fn(config):
    """Returns the learning decay function from options."""
    model_config = config.model
    running_config = config.training
    if running_config.decay_method == "noam":
        return functools.partial(
            noam_decay,
            warmup_steps=running_config.warmup_steps,
            model_size=model_config.hidden_size,
        )
    elif running_config.decay_method == "noamwd":
        return functools.partial(
            noamwd_decay,
            warmup_steps=running_config.warmup_steps,
            model_size=model_config.hidden_size,
            rate=running_config.learning_rate_decay,
            decay_steps=running_config.decay_steps,
            start_step=running_config.start_decay_steps,
        )
    elif running_config.decay_method == "cosine":
        return functools.partial(
            cosine_decay,
            warmup_steps=running_config.warmup_steps,
            train_steps=running_config.train_steps,
        )
    elif running_config.decay_method == "rsqrt":
        return functools.partial(rsqrt_decay, warmup_steps=running_config.warmup_steps)
    elif running_config.start_decay_steps is not None:
        return functools.partial(
            exponential_decay,
            rate=running_config.learning_rate_decay,
            decay_steps=running_config.decay_steps,
            start_step=running_config.start_decay_steps,
        )


def noam_decay(step, warmup_steps, model_size):
    """Learning rate schedule described in
    https://arxiv.org/pdf/1706.03762.pdf.
    """
    return model_size ** (-0.5) * min(step ** (-0.5), step * warmup_steps ** (-1.5))


def noamwd_decay(step, warmup_steps, model_size, rate, decay_steps, start_step=0):
    """Learning rate schedule optimized for huge batches"""
    return (
        model_size ** (-0.5)
        * min(step ** (-0.5), step * warmup_steps ** (-1.5))
        * rate ** (max(step - start_step + decay_steps, 0) // decay_steps)
    )


def cosine_decay(step, warmup_steps, train_steps):
    if step < warmup_steps:
        return step / warmup_steps
    else:
        decay_ratio = (step - warmup_steps) / (train_steps - warmup_steps)
        return 0.5 * (1.0 + cos(pi * decay_ratio))


def exponential_decay(step, rate, decay_steps, start_step=0):
    """A standard exponential decay, scaling the learning rate by :obj:`rate`
    every :obj:`decay_steps` steps.
    """
    return rate ** (max(step - start_step + decay_steps, 0) // decay_steps)


def rsqrt_decay(step, warmup_steps):
    """Decay based on the reciprocal of the step square root."""
    return 1.0 / sqrt(max(step, warmup_steps))


class MultipleOptimizer(object):
    """Implement multiple optimizers needed for sparse adam"""

    def __init__(self, op):
        """?"""
        self.optimizers = op

    @property
    def param_groups(self):
        param_groups = []
        for optimizer in self.optimizers:
            param_groups.extend(optimizer.param_groups)
        return param_groups

    def zero_grad(self, set_to_none=True):
        """?"""
        for op in self.optimizers:
            op.zero_grad(set_to_none)

    def step(self):
        """?"""
        for op in self.optimizers:
            op.step()

    @property
    def state(self):
        """?"""
        return {k: v for op in self.optimizers for k, v in op.state.items()}

    def state_dict(self):
        """?"""
        return [op.state_dict() for op in self.optimizers]

    def load_state_dict(self, state_dicts):
        """?"""
        assert len(state_dicts) == len(self.optimizers)
        for i in range(len(state_dicts)):
            self.optimizers[i].load_state_dict(state_dicts[i])


class Optimizer(object):
    """
    Controller class for optimization. Mostly a thin
    wrapper for `optim`, but also useful for implementing
    rate scheduling beyond what is currently available.
    Also implements necessary methods for training RNNs such
    as grad manipulations.

    Args:
        optimizer: A ``torch.optim.Optimizer`` instance.
        learning_rate: The initial learning rate.
        learning_rate_decay_fn: An optional callable taking the current step
            as argument and return a learning rate scaling factor.
        max_grad_norm: Clip gradients to this global norm.
    """

    def __init__(self, optimizer, learning_rate, learning_rate_decay_fn=None, max_grad_norm=None, use_amp=True):
        self._optimizer = optimizer
        self._learning_rate = learning_rate
        self._learning_rate_decay_fn = learning_rate_decay_fn
        self._max_grad_norm = max_grad_norm or 0
        self._training_step = 1
        self._decay_step = 1
        self.use_amp = use_amp
        self._scaler = None

    @classmethod
    def from_config(cls, model, config, checkpoint=None):
        """Builds the optimizer from options.

        Args:
          cls: The ``Optimizer`` class to instantiate.
          model: The model to optimize.
          opt: The dict of user options.
          checkpoint: An optional checkpoint to load states from.

        Returns:
          An ``Optimizer`` instance.
        """
        # we could almost go with only training config here, except for noam schedule which requires hidden size (additional kwarg?) # noqa: E501
        running_config = config.training
        optim_state_dict = None

        if running_config.train_from and checkpoint is not None and "optim" in checkpoint.keys():
            optim = checkpoint["optim"]
            ckpt_config = checkpoint["config"].training
            ckpt_state_dict = {}
            if isinstance(optim, Optimizer):  # Backward compatibility.
                ckpt_state_dict["training_step"] = optim._step + 1
                ckpt_state_dict["decay_step"] = optim._step + 1
                ckpt_state_dict["optimizer"] = optim.optimizer.state_dict()
            else:
                ckpt_state_dict = optim

            # we might be able to simplify this with the new general config update
            if running_config.reset_optim == "none":
                # Load everything from the checkpoint.
                running_config = ckpt_config
                optim_state_dict = ckpt_state_dict
            elif running_config.reset_optim == "all":
                # Build everything from scratch.
                pass
            elif running_config.reset_optim == "states":
                # Reset optimizer, keep options.
                running_config = ckpt_config
                optim_state_dict = ckpt_state_dict
                del optim_state_dict["optimizer"]
            elif running_config.reset_optim == "keep_states":
                # Reset options, keep optimizer.
                optim_state_dict = ckpt_state_dict

        use_amp = running_config.use_amp and running_config.compute_dtype in [torch.float16, torch.bfloat16]
        optimizer = cls(
            build_torch_optimizer(model, running_config),
            running_config.learning_rate,
            learning_rate_decay_fn=make_learning_rate_decay_fn(config),
            max_grad_norm=running_config.max_grad_norm,
            use_amp=use_amp,
        )
        # if running_config.compute_dtype in [torch.float16, torch.bfloat16]:
        if use_amp:
            optimizer._scaler = GradScaler("cuda")

        if optim_state_dict:
            optimizer.load_state_dict(optim_state_dict)
        return optimizer

    @property
    def training_step(self):
        """The current training step."""
        return self._training_step

    @property
    def amp(self):
        """True if use torch amp mix precision training."""
        return self.use_amp

    def learning_rate(self, step=None):
        """Returns the current learning rate."""
        if step is None:
            step = self._decay_step
        if self._learning_rate_decay_fn is None:
            return self._learning_rate
        scale = self._learning_rate_decay_fn(step)
        return scale * self._learning_rate

    def state_dict(self):
        return {
            "training_step": self._training_step,
            "decay_step": self._decay_step,
            "optimizer": self._optimizer.state_dict(),
        }

    def load_state_dict(self, state_dict):
        self._training_step = state_dict["training_step"]
        # State can be partially restored.
        if "decay_step" in state_dict:
            self._decay_step = state_dict["decay_step"]
        if "optimizer" in state_dict:
            self._optimizer.load_state_dict(state_dict["optimizer"])

    def zero_grad(self, set_to_none=True):
        """Zero the gradients of optimized parameters."""
        self._optimizer.zero_grad(set_to_none)

    def backward(self, loss):
        """Wrapper for backward pass. Some optimizer requires ownership of the
        backward pass."""
        if self._scaler is not None:
            self._scaler.scale(loss).backward()
        else:
            loss.backward()

    def step(self):
        """Update the model parameters based on current gradients.

        Optionally, will employ gradient modification or update learning
        rate.
        """
        learning_rate = self.learning_rate()

        if self._scaler is not None:
            self._scaler.unscale_(self._optimizer)

        for group in self._optimizer.param_groups:
            group["lr"] = learning_rate
            if self._max_grad_norm > 0:
                clip_grad_norm_(group["params"], self._max_grad_norm)

        if self._scaler is not None:
            # unscaled optimizer's gradients (already done therefore skip),
            # skips optimizer.step() if gradients contain infs/NaNs.
            self._scaler.step(self._optimizer)
            # Updates the scale for next iteration.
            self._scaler.update()
        else:
            self._optimizer.step()
        self._decay_step += 1
        self._training_step += 1


# Code below is an implementation of https://arxiv.org/pdf/1804.04235.pdf
# inspired but modified from https://github.com/DeadAt0m/adafactor-pytorch


class AdaFactor(torch.optim.Optimizer):
    def __init__(
        self,
        params,
        lr=None,
        beta1=0.9,
        beta2=0.999,
        eps1=1e-30,
        eps2=1e-3,
        cliping_threshold=1,
        non_constant_decay=True,
        enable_factorization=True,
        ams_grad=True,
        weight_decay=0,
    ):
        enable_momentum = beta1 != 0

        if non_constant_decay:
            ams_grad = False

        defaults = dict(
            lr=lr,
            beta1=beta1,
            beta2=beta2,
            eps1=eps1,
            eps2=eps2,
            cliping_threshold=cliping_threshold,
            weight_decay=weight_decay,
            ams_grad=ams_grad,
            enable_factorization=enable_factorization,
            enable_momentum=enable_momentum,
            non_constant_decay=non_constant_decay,
        )

        super(AdaFactor, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(AdaFactor, self).__setstate__(state)

    def _experimental_reshape(self, shape):
        temp_shape = shape[2:]
        if len(temp_shape) == 1:
            new_shape = (shape[0], shape[1] * shape[2])
        else:
            tmp_div = len(temp_shape) // 2 + len(temp_shape) % 2
            new_shape = (
                shape[0] * functools.reduce(operator.mul, temp_shape[tmp_div:], 1),
                shape[1] * functools.reduce(operator.mul, temp_shape[:tmp_div], 1),
            )
        return new_shape, copy(shape)

    def _check_shape(self, shape):
        """
        output1 - True - algorithm for matrix, False - vector;
        output2 - need reshape
        """
        if len(shape) > 2:
            return True, True
        elif len(shape) == 2:
            return True, False
        elif len(shape) == 2 and (shape[0] == 1 or shape[1] == 1):
            return False, False
        else:
            return False, False

    def _rms(self, x):
        return sqrt(torch.mean(x.pow(2)))

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data

                if grad.is_sparse:
                    raise RuntimeError(
                        "Adam does not support sparse \
                                       gradients, use SparseAdam instead"
                    )

                is_matrix, is_need_reshape = self._check_shape(grad.size())
                new_shape = p.data.size()
                if is_need_reshape and group["enable_factorization"]:
                    new_shape, old_shape = self._experimental_reshape(p.data.size())
                    grad = grad.view(new_shape)

                state = self.state[p]
                if len(state) == 0:
                    state["step"] = 0
                    if group["enable_momentum"]:
                        state["exp_avg"] = torch.zeros(new_shape, dtype=torch.float32, device=p.grad.device)

                    if is_matrix and group["enable_factorization"]:
                        state["exp_avg_sq_R"] = torch.zeros(
                            (1, new_shape[1]), dtype=torch.float32, device=p.grad.device
                        )
                        state["exp_avg_sq_C"] = torch.zeros(
                            (new_shape[0], 1), dtype=torch.float32, device=p.grad.device
                        )
                    else:
                        state["exp_avg_sq"] = torch.zeros(new_shape, dtype=torch.float32, device=p.grad.device)
                    if group["ams_grad"]:
                        state["exp_avg_sq_hat"] = torch.zeros(new_shape, dtype=torch.float32, device=p.grad.device)

                if group["enable_momentum"]:
                    exp_avg = state["exp_avg"]

                if is_matrix and group["enable_factorization"]:
                    exp_avg_sq_r = state["exp_avg_sq_R"]
                    exp_avg_sq_c = state["exp_avg_sq_C"]
                else:
                    exp_avg_sq = state["exp_avg_sq"]

                if group["ams_grad"]:
                    exp_avg_sq_hat = state["exp_avg_sq_hat"]

                state["step"] += 1
                lr_t = group["lr"]
                lr_t *= max(group["eps2"], self._rms(p.data))

                if group["enable_momentum"]:
                    if group["non_constant_decay"]:
                        beta1_t = (
                            group["beta1"]
                            * (1 - group["beta1"] ** (state["step"] - 1))
                            / (1 - group["beta1"] ** state["step"])
                        )
                    else:
                        beta1_t = group["beta1"]
                    exp_avg.mul_(beta1_t).add_(1 - beta1_t, grad)

                if group["non_constant_decay"]:
                    beta2_t = (
                        group["beta2"]
                        * (1 - group["beta2"] ** (state["step"] - 1))
                        / (1 - group["beta2"] ** state["step"])
                    )
                else:
                    beta2_t = group["beta2"]

                if is_matrix and group["enable_factorization"]:
                    exp_avg_sq_r.mul_(beta2_t).add_(
                        1 - beta2_t,
                        torch.sum(
                            torch.mul(grad, grad).add_(group["eps1"]),
                            dim=0,
                            keepdim=True,
                        ),
                    )
                    exp_avg_sq_c.mul_(beta2_t).add_(
                        1 - beta2_t,
                        torch.sum(
                            torch.mul(grad, grad).add_(group["eps1"]),
                            dim=1,
                            keepdim=True,
                        ),
                    )
                    v = torch.mul(exp_avg_sq_c, exp_avg_sq_r).div_(torch.sum(exp_avg_sq_r))
                else:
                    exp_avg_sq.mul_(beta2_t).addcmul_(1 - beta2_t, grad, grad).add_((1 - beta2_t) * group["eps1"])
                    v = exp_avg_sq

                g = grad
                if group["enable_momentum"]:
                    g = torch.div(exp_avg, 1 - beta1_t ** state["step"])

                if group["ams_grad"]:
                    torch.max(exp_avg_sq_hat, v, out=exp_avg_sq_hat)
                    v = exp_avg_sq_hat
                    u = torch.div(
                        g,
                        (torch.div(v, 1 - beta2_t ** state["step"])).sqrt().add_(group["eps1"]),
                    )
                else:
                    u = torch.div(g, v.sqrt())

                u.div_(max(1, self._rms(u) / group["cliping_threshold"]))
                p.data.add_(-lr_t * (u.view(old_shape) if is_need_reshape and group["enable_factorization"] else u))

                if group["weight_decay"] != 0:
                    p.data.add_(-group["weight_decay"] * lr_t, p.data)

        return loss



import codecs
import os
from eole.predict import GNMTGlobalScorer, Translator
from eole.config.run import (
    PredictConfig,
)  # probably should be done differently, but might work for now
from eole.constants import CorpusTask
from eole.inputters.dynamic_iterator import build_dynamic_dataset_iter
from eole.transforms import get_transforms_cls, make_transforms


class ScoringPreparator:
    """Allow the calculation of metrics via the Trainer's
    training_eval_handler method.
    """

    def __init__(self, vocabs, config):
        self.vocabs = vocabs
        self.config = config
        if self.config.dump_preds is not None:
            if not os.path.exists(self.config.dump_preds):
                os.makedirs(self.config.dump_preds)
        self.transforms = None
        self.transforms_cls = None

    def warm_up(self, transforms):
        self.transforms_cls = get_transforms_cls(transforms)
        self.transforms = make_transforms(self.config, self.transforms_cls, self.vocabs)

    def translate(self, model, gpu_rank, step):
        """Compute and save the sentences predicted by the
        current model's state related to a batch.

        Args:
            model (:obj:`eole.models.XXXModel`): The current model's state.
            transformed_batches(list of lists): A list of transformed batches.
            gpu_rank (int): Ordinal rank of the gpu where the
                translation is to be done.
            step: The current training step.
            mode: (string): 'train' or 'valid'.
        Returns:
            preds (list): Detokenized predictions
            texts_ref (list): Detokenized target sentences
        """
        # ########## #
        # Translator #
        # ########## #

        # Build translator from options
        model_config = self.config.model
        model_config._validate_model_config()

        # This is somewhat broken and we shall remove or improve
        # (take 'inference' field of config if exists?)
        # Set "default" translation options on empty cfgfile
        self.config.training.num_workers = 0
        predict_config = PredictConfig(
            model_path=["dummy"],
            src=self.config.data["valid"].path_src,
            compute_dtype=self.config.training.compute_dtype,
            beam_size=1,
            transforms=self.config.transforms,
            transforms_configs=self.config.transforms_configs,
            model=model_config,
            tgt_file_prefix=self.config.transforms_configs.prefix.tgt_prefix != "",
            gpu_ranks=[gpu_rank],
        )

        scorer = GNMTGlobalScorer.from_config(predict_config)
        translator = Translator.from_config(  # we need to review opt/config stuff in translator
            model,
            self.vocabs,
            predict_config,
            model_config,
            device_id=gpu_rank,
            global_scorer=scorer,
            report_align=predict_config.report_align,
            report_score=False,
            logger=None,
        )

        # ################### #
        # Validation iterator #
        # ################### #

        # Reinstantiate the validation iterator
        # Retrieve raw references and sources
        with codecs.open(self.config.data["valid"].path_tgt, "r", encoding="utf-8") as f:
            raw_refs = [line.strip("\n") for line in f if line.strip("\n")]
        with codecs.open(self.config.data["valid"].path_src, "r", encoding="utf-8") as f:
            raw_srcs = [line.strip("\n") for line in f if line.strip("\n")]

        infer_iter = build_dynamic_dataset_iter(
            predict_config,
            self.transforms,
            translator.vocabs,
            task=CorpusTask.INFER,
            tgt="",  # This force to clear the target side (needed when using tgt_file_prefix)
            device_id=gpu_rank,
        )

        # ########### #
        # Predictions #
        # ########### #
        _, _, preds = translator._predict(
            infer_iter,
            transform=infer_iter.transforms,
            attn_debug=predict_config.attn_debug,
            align_debug=predict_config.align_debug,
        )

        # ####### #
        # Outputs #
        # ####### #

        # Flatten predictions
        preds = [x.lstrip() for sublist in preds for x in sublist]
        # Save results
        if len(preds) > 0 and self.config.scoring_debug and self.config.dump_preds is not None:
            path = os.path.join(self.config.dump_preds, f"preds.valid_step_{step}.txt")
            with open(path, "a") as file:
                for i in range(len(raw_srcs)):
                    file.write("SOURCE: {}\n".format(raw_srcs[i]))
                    file.write("REF: {}\n".format(raw_refs[i]))
                    file.write("PRED: {}\n\n".format(preds[i]))
        return preds, raw_refs



"""Module that contain shard utils for dynamic data."""

import os
from eole.utils.logging import logger
from eole.constants import CorpusName, CorpusTask
from eole.transforms import TransformPipe
from eole.inputters.text_utils import transform_bucket

from contextlib import contextmanager
import itertools
import json


@contextmanager
def exfile_open(filename, *args, **kwargs):
    """Extended file opener enables open(filename=None).

    This context manager enables open(filename=None) as well as regular file.
    filename None will produce endlessly None for each iterate,
    while filename with valid path will produce lines as usual.

    Args:
        filename (str|None): a valid file path or None;
        *args: args relate to open file using codecs;
        **kwargs: kwargs relate to open file using codecs.

    Yields:
        `None` repeatly if filename==None,
        else yield from file specified in `filename`.
    """
    if filename is None:
        from itertools import repeat

        _file = repeat(None)
    else:
        import codecs

        _file = codecs.open(filename, *args, **kwargs)
    yield _file
    if filename is not None and _file:
        _file.close()


class BlockwiseCorpus(object):
    """A corpus class for reading a single file block by block."""

    def __init__(self, name, file_path, block_size=4096):
        """Initialize file path and block size."""
        self.id = name
        self.file_path = file_path
        self.block_size = block_size

    def load(self, offset=0, stride=1):
        """
        Load file and iterate by blocks.
        `offset` and `stride` allow iterating only on every
        `stride` block, starting from `offset`.
        """

        def make_ex(block_content):
            example = {
                "src": block_content,
                "tgt": block_content,
                "src_original": block_content,
                "tgt_original": block_content,
            }
            return example

        with open(self.file_path, mode="r", encoding="utf-8") as file:
            block_content = ""
            block_index = 0

            while True:
                chunk = file.read(self.block_size)
                if not chunk:
                    break

                if (block_index // stride) % stride == offset:
                    block_content += chunk

                    if len(chunk) < self.block_size:
                        # Reached end of file
                        yield make_ex(block_content)
                        break

                    if len(block_content) >= self.block_size:
                        yield make_ex(block_content)
                block_content = ""
                block_index += 1

    def __str__(self):
        cls_name = type(self).__name__
        return f"{cls_name}({self.id}, {self.file_path}, {self.file_path}" f"align={None}"


class ImageTextCorpus(object):
    """
    Handle vision data looking like this:
    ```
    [
        {
            "text": "List the top 5 countries in Europe with the highest GDP {image1}",
            "images": {
                "image1": "./test_data/gdp.png"
            }
        },
    ...
    ]
    ```
    """

    def __init__(self, name, data, is_train=False):
        self.id = name
        self.data = data
        self.is_train = is_train

    def load(self, offset=0, stride=1):
        def make_ex(item):
            example = {"text": item["text"], "images": item.get("images", {})}
            return example

        if isinstance(self.data, list):
            for i, item in enumerate(self.data):
                if (i // stride) % stride == offset:
                    yield make_ex(item)
        else:
            with exfile_open(self.data, mode="rb") as fs:
                data = json.load(fs)
                for i, item in enumerate(data):
                    if (i // stride) % stride == offset:
                        yield make_ex(item)

    def __str__(self):
        cls_name = type(self).__name__
        return f"{cls_name}({self.id}, {self.path}"


class ParallelCorpus(object):
    """A parallel corpus file pair that can be loaded to iterate."""

    def __init__(self, name, src, tgt, sco=None, align=None):
        """Initialize src & tgt side file path."""
        self.id = name
        self.src = src
        self.tgt = tgt
        self.sco = sco
        self.align = align

    def _load_hf_dataset(self, hf_string):
        """
        Load a Hugging Face dataset from the given identifier.
        Matchs the 'hf://' prefix and assumes the dataset is in streaming mode.
        Match the last '/field' to get the language / score field
        Potentially a config_name field between dataset_name and last field
        """
        from datasets import load_dataset

        parts = hf_string.replace("hf://", "").split("/")
        dataset_name = "/".join(parts[:2])
        remaining_parts = parts[2:]

        if len(remaining_parts) == 1:
            return load_dataset(dataset_name, split="train", streaming=True)
        elif len(remaining_parts) == 2:
            return load_dataset(dataset_name, remaining_parts[0], split="train", streaming=True)

    def load(self, offset=0, stride=1):
        """
        Load file and iterate by lines.
        `offset` and `stride` allow to iterate only on every
        `stride` example, starting from `offset`.
        In the case of local files, all files are open exactly the same way by each worker
        Therefore we need to apply a stride / offset rule to make sure we do not process the same ex.
        In the case of HF streaming mode we need to make sure we have more shards than workers.
        Typically we recommend to have shard being a multiple of workers for instance for big datasets:
        16 shards for 4 workers. The shards will be iterated automatically since HF locks shards when in use.
        """

        def make_ex(sline, tline, scoline, align):
            # 'src_original' and 'tgt_original' store the
            # original line before tokenization.
            example = {
                "src": sline,
                "tgt": tline,
                "sco": scoline,
                "src_original": sline,
                "tgt_original": tline,
            }
            if align is not None:
                example["align"] = align
            return example

        if isinstance(self.src, list):
            fs = self.src
            ft = [] if self.tgt is None else self.tgt
            fsco = [] if self.sco is None else self.sco
            fa = [] if self.align is None else self.align
            for i, (sline, tline, scoline, align) in enumerate(itertools.zip_longest(fs, ft, fsco, fa)):
                if (i // stride) % stride == offset:
                    if scoline is None:
                        scoline = 1.0
                    yield make_ex(sline, tline, scoline, align)

        elif self.src.startswith("hf://"):
            # If `src` is a Hugging Face dataset identifier
            dataset = self._load_hf_dataset(self.src)
            for i, example in enumerate(dataset):
                sline = example.get(self.src.split("/")[-1])
                tline = example.get(self.tgt.split("/")[-1]) if self.tgt is not None else None
                if self.sco is not None:
                    scoline = example.get(self.sco.split("/")[-1], 1.0)
                else:
                    scoline = 1.0
                yield make_ex(sline, tline, scoline, None)

        else:
            with exfile_open(self.src, mode="rb") as fs, exfile_open(self.tgt, mode="rb") as ft, exfile_open(
                self.sco, mode="rb"
            ) as fsco, exfile_open(self.align, mode="rb") as fa:
                for i, (sline, tline, scoline, align) in enumerate(zip(fs, ft, fsco, fa)):
                    if (i // stride) % stride == offset:
                        if tline is not None:
                            tline = tline.decode("utf-8")
                        if scoline is not None:
                            scoline = float(scoline.decode("utf-8"))
                        else:
                            scoline = 1.0
                        if align is not None:
                            align = align.decode("utf-8")
                        yield make_ex(sline.decode("utf-8"), tline, scoline, align)

    def __str__(self):
        cls_name = type(self).__name__
        return f"{cls_name}({self.id}, {self.src}, {self.tgt}, " f"align={self.align}"


def get_corpora(config, task=CorpusTask.TRAIN, src=None, tgt=None, align=None):
    corpora_dict = {}
    if task == CorpusTask.TRAIN:
        for corpus_id, corpus_dict in config.data.items():
            if corpus_id != CorpusName.VALID:
                if corpus_dict.path_txt is None:
                    corpora_dict[corpus_id] = ParallelCorpus(
                        corpus_id,
                        corpus_dict.path_src,
                        corpus_dict.path_tgt,
                        corpus_dict.path_sco,
                        corpus_dict.path_align,
                    )
                elif config.data_type == "text":
                    corpora_dict[corpus_id] = BlockwiseCorpus(
                        corpus_id,
                        corpus_dict.path_txt,
                        block_size=8192,  # number of characters
                    )
                elif config.data_type == "image":
                    corpora_dict[corpus_id] = ImageTextCorpus(
                        corpus_id,
                        corpus_dict.path_txt,
                        is_train=True,
                    )
    elif task == CorpusTask.VALID:
        if CorpusName.VALID in config.data.keys():
            if config.data_type == "text":
                corpora_dict[CorpusName.VALID] = ParallelCorpus(
                    CorpusName.VALID,
                    config.data[CorpusName.VALID].path_src,
                    config.data[CorpusName.VALID].path_tgt,
                    None,
                    config.data[CorpusName.VALID].path_align,
                )
            elif config.data_type == "image":
                corpora_dict[CorpusName.VALID] = ImageTextCorpus(
                    CorpusName.VALID,
                    config.data[CorpusName.VALID].path_txt,
                    is_train=True,
                )
        else:
            return None
    else:
        if config.data_type == "text":
            corpora_dict[CorpusName.INFER] = ParallelCorpus(
                CorpusName.INFER,
                src if src else config.src,
                tgt if tgt else config.tgt,
                align if align else None,
            )
        elif config.data_type == "image":
            corpora_dict[CorpusName.INFER] = ImageTextCorpus(
                CorpusName.INFER, src, is_train=False  # maybe homogenize to some better name
            )
    return corpora_dict


class ParallelCorpusIterator(object):
    """An iterator dedicated to ParallelCorpus.

    Args:
        corpus (ParallelCorpus): corpus to iterate;
        transform (TransformPipe): transforms to be applied to corpus;
        skip_empty_level (str): security level when encouter empty line;
        stride (int): iterate corpus with this line stride;
        offset (int): iterate corpus with this line offset.
    """

    def __init__(self, corpus, transform, skip_empty_level="warning", stride=1, offset=0, is_train=False):
        self.cid = corpus.id
        self.corpus = corpus
        self.transform = transform
        if skip_empty_level not in ["silent", "warning", "error"]:
            raise ValueError(f"Invalid argument skip_empty_level={skip_empty_level}")
        self.skip_empty_level = skip_empty_level
        self.stride = stride
        self.offset = offset

    def _process(self, stream):
        for i, example in enumerate(stream):
            example["src"] = example["src"].strip().split(" ")
            example["src_original"] = example["src_original"].strip().split(" ")
            line_number = i * self.stride + self.offset
            example["cid_line_number"] = line_number
            example["cid"] = self.cid
            if "align" in example:
                example["align"] = example["align"].strip().split(" ")
            if example["tgt"] is not None:
                example["tgt"] = example["tgt"].strip().split(" ")
                example["tgt_original"] = example["tgt_original"].strip().split(" ")
                if (
                    len(example["src"]) == 0
                    or len(example["tgt"]) == 0
                    or ("align" in example and example["align"] == 0)
                ):
                    # empty example: skip
                    empty_msg = f"Empty line  in {self.cid}#{line_number}."
                    if self.skip_empty_level == "error":
                        raise IOError(empty_msg)
                    elif self.skip_empty_level == "warning":
                        logger.warning(empty_msg)
                    if len(example["src"]) == 0 and len(example["tgt"]) == 0:
                        yield (example, self.transform, self.cid)
                    continue
            yield (example, self.transform, self.cid)
        report_msg = self.transform.stats()
        if report_msg != "":
            logger.info("* Transform statistics for {}({:.2f}%):\n{}\n".format(self.cid, 100 / self.stride, report_msg))

    def __iter__(self):
        corpus_stream = self.corpus.load(stride=self.stride, offset=self.offset)
        corpus = self._process(corpus_stream)
        yield from corpus


class ImageTextCorpusIterator(object):
    def __init__(
        self,
        corpus,
        transform,
        skip_empty_level="warning",
        stride=1,
        offset=0,
        is_train=False,
        image_patch_size=16,
        image_size=1024,
        adapter="llava",
    ):
        self.cid = corpus.id
        self.corpus = corpus
        self.transform = transform
        if skip_empty_level not in ["silent", "warning", "error"]:
            raise ValueError(f"Invalid argument skip_empty_level={skip_empty_level}")
        self.skip_empty_level = skip_empty_level
        self.stride = stride
        self.offset = offset
        self.is_train = is_train
        self.image_patch_size = image_patch_size
        self.image_size = image_size
        self.adapter = adapter

    def _process(self, stream):
        from eole.inputters.image_utils import process_image

        for i, example in enumerate(stream):
            # process images
            processed_images = {
                k: process_image(
                    v, adapter=self.adapter, image_patch_size=self.image_patch_size, image_size=self.image_size
                )
                for k, v in example.get("images", {}).items()
            }
            text = example["text"]
            image_tokens = {
                # using a list here would induce unwanted whitespaces between tokens downstream
                k: "".join(v["tokens"])
                for k, v in processed_images.items()
            }
            text = text.format(**image_tokens)
            line_number = i * self.stride + self.offset
            example = {
                "src": text,
                "tgt": text if self.is_train else None,
                "images": {k: v["image"] for k, v in processed_images.items()},
                "cid": self.cid,
                "cid_line_number": line_number,
            }
            yield example, self.transform, self.cid

    def __iter__(self):
        corpus_stream = self.corpus.load(stride=self.stride, offset=self.offset)
        corpus = self._process(corpus_stream)
        yield from corpus


def build_corpora_iters(
    corpora,
    transforms,
    corpora_info,
    skip_empty_level="warning",
    stride=1,
    offset=0,
    image_patch_size=16,
    image_size=1024,
    adapter="llava",
):
    """Return `ParallelCorpusIterator` for all corpora defined in opts."""
    corpora_iters = dict()
    for c_id, corpus in corpora.items():
        transform_names = corpora_info[c_id].transforms
        corpus_transform = [transforms[name] for name in transform_names if name in transforms]
        transform_pipe = TransformPipe.build_from(corpus_transform)
        if isinstance(corpus, ParallelCorpus):
            corpus_iter = ParallelCorpusIterator(
                corpus,
                transform_pipe,
                skip_empty_level=skip_empty_level,
                stride=stride,
                offset=offset,
                is_train=getattr(corpus, "is_train", None),
            )
        elif isinstance(corpus, ImageTextCorpus):
            corpus_iter = ImageTextCorpusIterator(
                corpus,
                transform_pipe,
                skip_empty_level=skip_empty_level,
                stride=stride,
                offset=offset,
                is_train=getattr(corpus, "is_train", None),
                image_patch_size=image_patch_size,
                image_size=image_size,
                adapter=adapter,
            )
        corpora_iters[c_id] = corpus_iter
    return corpora_iters


def save_transformed_sample(config, transforms, n_sample=3):
    """Save transformed data sample as specified in opts."""

    if n_sample == -1:
        logger.info(f"n_sample={n_sample}: Save full transformed corpus.")
    elif n_sample == 0:
        logger.info(f"n_sample={n_sample}: no sample will be saved.")
        return
    elif n_sample > 0:
        logger.info(f"Save {n_sample} transformed example/corpus.")
    else:
        raise ValueError(f"n_sample should >= -1, get {n_sample}.")

    corpora = get_corpora(config, CorpusTask.TRAIN)
    datasets_iterables = build_corpora_iters(corpora, transforms, config.data, skip_empty_level=config.skip_empty_level)
    sample_path = os.path.join(config.save_data, "samples", CorpusName.SAMPLE)
    os.makedirs(sample_path, exist_ok=True)
    for c_name, c_iter in datasets_iterables.items():
        dest_base = os.path.join(sample_path, "{}.{}".format(c_name, CorpusName.SAMPLE))
        with open(dest_base + ".src", "w", encoding="utf-8") as f_src, open(
            dest_base + ".tgt", "w", encoding="utf-8"
        ) as f_tgt:
            bucket = []
            for i, ex in enumerate(c_iter):
                if i > n_sample:
                    break
                else:
                    bucket.append(ex)
            pro_bucket = transform_bucket(CorpusTask.TRAIN, bucket)
            if pro_bucket is not None:
                for maybe_example in pro_bucket:
                    if maybe_example is not None:
                        src_line, tgt_line = (
                            maybe_example["src"]["src"],
                            maybe_example["tgt"]["tgt"],
                        )

                        f_src.write(src_line + "\n")
                        f_tgt.write(tgt_line + "\n")



#!/usr/bin/env python
import json
import torch
import pyonmttok
from eole.inputters.inputter import vocabs_to_dict
import os
from safetensors.torch import save_file

from eole.config.models import (
    EmbeddingsConfig,
    TransformerLMModelConfig,
)
from eole.config.run import TrainConfig
from eole.config.training import TrainingConfig
from eole.config import recursive_model_fields_set
from eole.bin import BaseBin, register_bin


@register_bin(name="redpajama")
class RedPajamaConverter(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--model_dir",
            type=str,
            required=True,
            help="""Path to the model directory""",
        )
        parser.add_argument(
            "--vocab_file",
            type=str,
            required=True,
            help="""Path to the tokenizer model""",
        )
        parser.add_argument("--output", type=str, required=True, help="""Path to the model directory""")
        parser.add_argument("--nshards", type=int, default=1, help="""Path to the model directory""")

    @classmethod
    def run(cls, args):
        print("Loading the full model in Cpu Ram ...")
        from transformers import AutoModelForCausalLM

        model = AutoModelForCausalLM.from_pretrained(
            args.model_dir,
            torch_dtype=torch.float16,
            device_map={"": "cpu"},
            trust_remote_code=True,
        )
        checkpoint = model.state_dict()

        params_json = os.path.join(args.model_dir, "config.json")
        with open(params_json, encoding="utf-8") as fparam:
            params = json.load(fparam)

        decoder_layers = params["num_hidden_layers"]
        src_word_vec_size = params["hidden_size"]
        tgt_word_vec_size = params["hidden_size"]
        hidden_size = params["hidden_size"]
        heads = params["num_attention_heads"]
        vocab_size = params["vocab_size"]
        transformer_ff = params["intermediate_size"]

        for shard in range(args.nshards):

            print("starting output shard: %d/%d" % (shard + 1, args.nshards))
            eole_safetensor = {}

            if shard == 0:
                eole_safetensor["tgt_emb.embeddings.weight"] = checkpoint["gpt_neox.embed_in.weight"]

                eole_safetensor["decoder.layer_norm.weight"] = checkpoint["gpt_neox.final_layer_norm.weight"]
                eole_safetensor["decoder.layer_norm.bias"] = checkpoint["gpt_neox.final_layer_norm.bias"]

                eole_safetensor["generator.weight"] = checkpoint["embed_out.weight"]
                eole_safetensor["generator.bias"] = torch.zeros(
                    eole_safetensor["generator.weight"].size(0), dtype=torch.float16
                )

            for i in range(
                -(decoder_layers // -args.nshards) * shard,
                min(-(decoder_layers // -args.nshards) * (shard + 1), decoder_layers),
                1,
            ):
                # redpajama stores QKV in one single tensor but it is not simply piled up Q+K+V
                # it is heads interleaved to we need to slice first
                # also it uses the HF rotary so we need to permute Q and K interleave

                qkv_W = checkpoint["gpt_neox.layers." + str(i) + ".attention.query_key_value.weight"].view(
                    heads, 3 * hidden_size // heads, hidden_size
                )
                qkv_B = checkpoint["gpt_neox.layers." + str(i) + ".attention.query_key_value.bias"].view(
                    heads, 3 * hidden_size // heads
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_query.weight"] = (
                    qkv_W[:, : hidden_size // heads, :]
                    .view(heads, 2, hidden_size // heads // 2, hidden_size)
                    .transpose(1, 2)
                    .reshape(hidden_size, hidden_size)
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_keys.weight"] = (
                    qkv_W[:, hidden_size // heads : 2 * hidden_size // heads, :]
                    .view(heads, 2, hidden_size // heads // 2, hidden_size)
                    .transpose(1, 2)
                    .reshape(hidden_size, hidden_size)
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_values.weight"] = qkv_W[
                    :, 2 * hidden_size // heads : 3 * hidden_size // heads, :
                ].reshape(hidden_size, hidden_size)

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_query.bias"] = (
                    qkv_B[:, : hidden_size // heads]
                    .view(heads, 2, hidden_size // heads // 2)
                    .transpose(1, 2)
                    .reshape(hidden_size)
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_keys.bias"] = (
                    qkv_B[:, hidden_size // heads : 2 * hidden_size // heads]
                    .view(heads, 2, hidden_size // heads // 2)
                    .transpose(1, 2)
                    .reshape(hidden_size)
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_values.bias"] = qkv_B[
                    :, 2 * hidden_size // heads : 3 * hidden_size // heads
                ].reshape(hidden_size)

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.final_linear.weight"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".attention.dense.weight"
                ]
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.final_linear.bias"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".attention.dense.bias"
                ]

                eole_safetensor["decoder.transformer_layers." + str(i) + ".layer_norm_1.weight"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".input_layernorm.weight"
                ]
                eole_safetensor["decoder.transformer_layers." + str(i) + ".layer_norm_1.bias"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".input_layernorm.bias"
                ]

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_1.weight"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".mlp.dense_h_to_4h.weight"
                ]
                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_1.bias"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".mlp.dense_h_to_4h.bias"
                ]

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_2.weight"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".mlp.dense_4h_to_h.weight"
                ]
                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_2.bias"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".mlp.dense_4h_to_h.bias"
                ]

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.layer_norm.weight"] = (
                    checkpoint["gpt_neox.layers." + str(i) + ".post_attention_layernorm.weight"]
                )
                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.layer_norm.bias"] = checkpoint[
                    "gpt_neox.layers." + str(i) + ".post_attention_layernorm.bias"
                ]

            if shard == 0:
                transformer_ff = eole_safetensor["decoder.transformer_layers.0.feed_forward.w_1.weight"].size(0)
                vocab_size = eole_safetensor["generator.weight"].size(0)
            print("Saving output model shard: %d" % shard)
            save_file(
                eole_safetensor,
                os.path.join(args.output, "model.{:02d}.safetensors".format(shard)),
            )

        vocabs = {}
        with open(args.vocab_file, "r", encoding="utf-8") as vocab:
            src_vocab = pyonmttok.build_vocab_from_lines(vocab)

        vocabs["src"] = src_vocab
        vocabs["tgt"] = src_vocab
        vocabs["decoder_start_token"] = "</s>"

        vocab_dict = vocabs_to_dict(vocabs)
        with open(os.path.join(args.output, "vocab.json"), "w", encoding="utf-8") as f:
            json.dump(vocab_dict, f, indent=2, ensure_ascii=False)

        with open(os.path.join(args.output, "vocab.txt"), "w", encoding="utf-8") as vocabfile:
            for tok in vocab_dict["src"]:
                vocabfile.write(tok + "\n")

        position_encoding = {
            "position_encoding_type": "Rotary",
            "n_positions": 0,
        }

        config = TrainConfig(
            data=None,
            skip_empty_level="silent",  # default is "warning"
            save_data=None,
            n_sample=0,
            src_vocab=None,
            tgt_vocab=None,
            share_vocab=True,
            src_vocab_size=vocab_size,
            tgt_vocab_size=vocab_size,
            vocab_size_multiple=8,
            decoder_start_token=vocabs["decoder_start_token"],
            transforms=["filtertoolong"],
            transforms_configs={"filtertoolong": {"src_seq_length": 512, "tgt_seq_length": 512}},
            model=TransformerLMModelConfig(
                layers=decoder_layers,
                hidden_size=hidden_size,
                heads=heads,
                transformer_ff=transformer_ff,
                embeddings=EmbeddingsConfig(
                    src_word_vec_size=src_word_vec_size,
                    tgt_word_vec_size=tgt_word_vec_size,
                    **position_encoding,
                ),
                # src_word_vec_size=src_word_vec_size,
                # tgt_word_vec_size=tgt_word_vec_size,
                layer_norm="standard",
                pos_ffn_activation_fn="gelu",
                parallel_residual=False,
                add_qkvbias=True,
                add_ffnbias=True,
            ),
            training=TrainingConfig(
                compute_dtype="fp16",
                batch_size=896,
                batch_size_multiple=1,
                batch_type="tokens",
                normalization="tokens",
                accum_count=[32],
                accum_steps=[0],
                valid_batch_size=256,
            ),
        )

        config_dict = recursive_model_fields_set(config)
        with open(os.path.join(args.output, "config.json"), "w", encoding="utf-8") as f:
            json.dump(config_dict, f, indent=2, ensure_ascii=False)



# converts a SentencePiece vocabulary to the format expected by dynamic data
# (essentially converts float expected counts to "fixed precision" int pseudo
# counts)
import sys
import math
from eole.constants import DefaultTokens

from eole.bin import BaseBin, register_bin


@register_bin(name="spm_to_vocab")
class SpmtoVocab(BaseBin):
    OMIT = (DefaultTokens.UNK, DefaultTokens.BOS, DefaultTokens.EOS)

    @classmethod
    def add_args(cls, parser):
        # legacy tool directly reads lines from stdin
        pass

    @classmethod
    def convert(cls, lines):
        for line in lines:
            w, c = line.rstrip("\n").split(None, 1)
            if w in cls.OMIT:
                continue
            c = math.exp(float(c)) * 1000000
            c = int(c) + 1
            yield w, c

    @classmethod
    def run(cls, args):
        for c, w in cls.convert(sys.stdin):
            print("{}\t{}".format(c, w))



# Main Entrypoint

from argparse import ArgumentParser
from eole.bin import AVAILABLE_BINS


def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(help="Executable to run.", dest="bin")
    for bin_type, bins in AVAILABLE_BINS.items():
        # specific case for "run" which should be done at root level
        if bin_type == "run":
            for bin_name, bin_cls in bins.items():
                subparser = subparsers.add_parser(bin_name)
                bin_cls.add_args(subparser)

        else:
            subparser = subparsers.add_parser(bin_type)
            sub_subparsers = subparser.add_subparsers(dest="sub_bin")
            for bin_name, bin_cls in bins.items():
                sub_subparser = sub_subparsers.add_parser(bin_name)  # define some helpstring in the tool class
                bin_cls.add_args(sub_subparser)

    args = parser.parse_args()

    if args.bin in AVAILABLE_BINS.keys() and args.sub_bin in AVAILABLE_BINS[args.bin].keys():
        bin_cls = AVAILABLE_BINS[args.bin][args.sub_bin]
    elif args.bin in AVAILABLE_BINS["run"].keys():
        bin_cls = AVAILABLE_BINS["run"][args.bin]
    else:
        print(parser.format_help())
        exit(1)

    bin_cls.run(args)


if __name__ == "__main__":
    main()



from eole.utils.logging import logger
from eole.transforms import register_transform
from .transform import Transform, TransformConfig
from pydantic import Field
import numpy as np
import time


class FuzzyMatchConfig(TransformConfig):
    tm_path: str | None = Field(default=None, description="Path to a flat text TM.")
    fuzzy_corpus_ratio: float | None = Field(default=0.1, description="Ratio of corpus to augment with fuzzy matches.")
    fuzzy_threshold: float | None = Field(default=70, description="The fuzzy matching threshold.")
    tm_delimiter: str | None = Field(default="\t", description="The delimiter used in the flat text TM.")
    fuzzy_token: str | None = Field(default="｟fuzzy｠", description="The fuzzy token to be added with the matches.")
    fuzzymatch_min_length: int | None = Field(default=4, description="Min length for TM entries and examples to match.")
    fuzzymatch_min_length: int | None = Field(
        default=70, description="Max length for TM entries and examples to match."
    )


class FuzzyMatcher(object):
    """Class for creating and setting up fuzzy matchers."""

    def __init__(
        self,
        tm_path,
        corpus_ratio,
        threshold=70,
        tm_delimiter="\t",
        fuzzy_token="｟fuzzy｠",
        tm_unit_min_lentgh=4,
        tm_unit_max_length=70,
    ):
        self.threshold = threshold
        self.corpus_ratio = corpus_ratio
        self.tm_delimiter = tm_delimiter
        self.fuzzy_token = fuzzy_token
        self.tm_unit_min_length = tm_unit_min_lentgh
        self.tm_unit_max_length = tm_unit_max_length
        self.internal_tm = self._create_tm(tm_path)

    def _create_tm(self, tm_path):
        """The TM should be a utf-8 text file with each line
        containing a source sentence and its translation, separated
        by the `self.tm_delimiter`. A TM size of 200k-250k pairs should
        provide enough matches and good performance, but this may
        depend on overall system specs (RAM, CPU)
        """

        src_segments, tgt_segments = list(), list()
        with open(tm_path, mode="r", encoding="utf-8") as file:
            pairs = file.readlines()
            for pair in pairs:
                source, target = map(str, pair.split(self.tm_delimiter))

                # Filter out very short or very long sentences
                # from the TM for better performance
                if len(source) < self.tm_unit_min_length or len(source) > self.tm_unit_max_length:
                    continue
                src_segments.append(source.strip())
                tgt_segments.append(target.strip())
        logger.debug(f"Translation Memory size for FuzzyMatch transform: " f"{len(src_segments)}")
        return [src_segments, tgt_segments]

    def _get_batch_matches(self, batch):
        from rapidfuzz import fuzz, process

        logger.debug(f"Starting fuzzy matching on {len(batch)} examples")
        fuzzy_count = 0
        start = time.time()
        augmented = list()

        # We split the `batch` and perform fuzzy matching
        # in smaller chunks of 10.000 examples in order to
        # reduce memory usage.
        # Perfomance is not affected.
        chunk_size = 10000
        mini_batches = np.array_split(batch, len(batch) // chunk_size if len(batch) > chunk_size else 1)
        for mini_batch in mini_batches:
            plist = list(mini_batch)
            if fuzzy_count >= len(batch) * self.corpus_ratio:
                augmented.extend(plist)
                continue

            results = process.cdist(
                plist,
                self.internal_tm[0],
                scorer=fuzz.ratio,
                dtype=np.uint8,
                score_cutoff=self.threshold,
                workers=-1,
            )

            matches = np.any(results, 1)
            argmax = np.argmax(results, axis=1)
            for idx, s in enumerate(plist):
                # Probably redundant but let's be safe
                # in case some examples are already fuzzied
                # (e.g. from another pipeline or workflow)
                if self.fuzzy_token in s:
                    continue
                # We don't want exact matches
                if matches[idx] and results[idx][argmax[idx]] < 100:
                    if fuzzy_count >= len(batch) * self.corpus_ratio:
                        break
                    plist[idx] = s + self.fuzzy_token + self.internal_tm[1][argmax[idx]]
                    fuzzy_count += 1
            augmented.extend(plist)

        end = time.time()
        logger.debug(f"FuzzyMatch Transform: Added {fuzzy_count} " f"fuzzies in {end-start} secs")

        return augmented


@register_transform(name="fuzzymatch")
class FuzzyMatchTransform(Transform):
    """Perform fuzzy matching against a translation memory and
    augment source examples with target matches for Neural Fuzzy Repair.
    :cite:`bulte-tezcan-2019-neural`
    """

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.tm_path = self.config.tm_path
        self.fuzzy_corpus_ratio = self.config.fuzzy_corpus_ratio
        self.fuzzy_threshold = self.config.fuzzy_threshold
        self.tm_delimiter = self.config.tm_delimiter
        self.fuzzy_token = self.config.fuzzy_token
        self.fuzzymatch_min_length = self.config.fuzzymatch_min_length
        self.fuzzymatch_max_length = self.config.fuzzymatch_max_length

    @classmethod
    def get_specials(cls, config):
        """Add the fuzzy match token to the src vocab."""

        return ([config.fuzzy_token], list())

    def warm_up(self, vocabs=None):
        """Create the fuzzy matcher."""

        super().warm_up(None)
        self.matcher = FuzzyMatcher(
            self.tm_path,
            self.fuzzy_corpus_ratio,
            self.fuzzy_threshold,
            self.tm_delimiter,
            self.fuzzy_token,
            self.fuzzymatch_min_length,
            self.fuzzymatch_max_length,
        )

    def apply(self, example, is_train=False, stats=None, **kwargs):
        return example

    def batch_apply(self, batch, is_train=False, stats=None, **kwargs):
        src_segments = list()
        for ex, _, _ in batch:
            # Apply a basic filtering to leave out very short or very long
            # sentences and speed up things a bit during fuzzy matching
            if (
                len(" ".join(ex["src"])) > self.fuzzymatch_min_length
                and len(" ".join(ex["src"])) < self.fuzzymatch_max_length
            ):
                src_segments.append(" ".join(ex["src"]))
            else:
                src_segments.append("")
        fuzzied_src = self.matcher._get_batch_matches(src_segments)
        assert len(src_segments) == len(fuzzied_src)
        for idx, (example, _, _) in enumerate(batch):
            if fuzzied_src[idx] != "":
                example["src"] = fuzzied_src[idx].split(" ")

        return batch



from eole.bin import BaseBin, register_bin
import yaml
from eole.config.run import TrainConfig
from eole.config.training import TrainingConfig
from eole.config.run import NestedAllTransformsConfig
from collections import defaultdict, OrderedDict
from eole.config.models import BaseModelConfig
from rich import print
from eole.config import reorder_fields


# to be extended
KEY_MAP = {
    "save_model": ["training", "model_path"],
    "encoder_type": [
        "model",
        "encoder",
        "encoder_type",
    ],  # will set config["model"]["encoder"]["encoder_type"]
    "decoder_type": ["model", "decoder", "decoder_type"],
    "enc_layers": ["model", "encoder", "layers"],
    "dec_layers": ["model", "decoder", "layers"],
    "heads": ["model", "heads"],
    "model_dtype": ["training", "compute_dtype"],
    "max_relative_positions": ["model", "embeddins", "n_positions"],
    "num_kv": ["model", "heads_kv"],
    "pos_ffn_activation_fn": ["model", "mlp_activation_fn"],
}


def custom_mapping(k, v, config):
    # to be extended
    match k, v:
        case "max_relative_positions", -1:
            config["model"]["embeddings"]["position_encoding_type"] = "Rotary"
        case "max_relative_positions", -2:
            config["model"]["embeddings"]["position_encoding_type"] = "Alibi"
        case "max_relative_positions", _ if v > 0:
            config["model"]["embeddings"]["positions_encoding_type"] = "Relative"
        case "max_relative_positions", 0:
            config["model"]["embeddings"]["position_encoding_type"] = "Absolute"


def set_nested_value(d, keys, value):
    current = d
    for key in keys[:-1]:
        if key not in current:
            current[key] = {}
        current = current[key]
    current[keys[-1]] = value


def _to_dict(d):
    if isinstance(d, defaultdict) or isinstance(d, OrderedDict):
        d = {k: _to_dict(v) for k, v in d.items()}
    return d


@register_bin(name="onmt_config")
class OnmtConfigConverter(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument("input", help="Input OpenNMT-py yaml config to convert.")
        parser.add_argument("output", help="Output converted yaml EOLE config.")
        parser.add_argument(
            "-a",
            "--architecture",
            choices=["rnn", "cnn", "transformer", "transformer_lm"],
            help="Setting this field will ensure conversion will be relevant to the architecture.",
        )

    @classmethod
    def run(cls, args):
        with open(args.input) as f:
            data_in = yaml.safe_load(f)
        # 3 potential levels of nesting
        new_config = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
        if "data" in data_in.keys():
            new_config["data"] = data_in["data"]
        if args.architecture is not None:
            new_config["model"]["architecture"] = args.architecture
        mapped = {"data"}
        # retrieve all transforms to populate only necessary transforms_configs
        all_transforms = set(data_in.get("transforms", []))
        for _, corpus in data_in.get("data", {}).items():
            all_transforms |= set(corpus.get("transforms", []))
        for k, v in data_in.items():
            if k in KEY_MAP.keys():
                set_nested_value(new_config, KEY_MAP[k], v)
                mapped.add(k)
            if k not in mapped:
                if k in TrainConfig.model_fields:
                    new_config[k] = v
                    mapped.add(k)
                elif k in TrainingConfig.model_fields:
                    new_config["training"][k] = v
                    mapped.add(k)
                elif k in BaseModelConfig.model_fields:
                    new_config["model"][k] = v
                    mapped.add(k)
                else:
                    for (
                        t_name,
                        t_field,
                    ) in NestedAllTransformsConfig.model_fields.items():
                        if k in t_field.default.model_fields:
                            if t_name in all_transforms:
                                try:
                                    # handle edge cases like {src,tgt}_onmttok_kwargs
                                    # str to dict conversion
                                    v = eval(v)
                                except Exception:
                                    pass
                                new_config["transforms_configs"][t_name][k] = v
                                mapped.add(k)
        unmapped = set(data_in.keys()) - mapped
        # reorder fields for readability
        new_config = reorder_fields(new_config)
        # convert to standard dict for proper yaml dump
        new_config = _to_dict(new_config)
        print(new_config)
        print("Saving converted config to:", args.output)
        with open(args.output, "w") as f:
            yaml.dump(new_config, f, default_flow_style=False, sort_keys=False)
        print("Remaining unmapped items:", unmapped)
        # test config validity
        new_config["data"] = {}
        try:
            TrainConfig(**new_config)
        except Exception as e:
            print(e)
        else:
            print("Converted config seems ok!")



#!/usr/bin/env python
"""Get vocabulary coutings from transformed corpora samples."""
import os
import copy
import multiprocessing as mp
import pyonmttok
from functools import partial
from eole.utils.logging import init_logger, logger
from eole.utils.misc import set_random_seed, check_path
from argparse import ArgumentParser

from eole.inputters.text_corpus import build_corpora_iters, get_corpora
from eole.inputters.text_utils import transform_bucket
from eole.transforms import make_transforms, get_transforms_cls
from eole.constants import CorpusName, CorpusTask
from collections import Counter

from eole.config.run import BuildVocabConfig
from eole.bin import register_bin
from eole.bin.run import RunBin


MAXBUCKETSIZE = 256000


def write_files_from_queues(sample_path, queues):
    """
    Standalone process that reads data from
    queues in order and write to sample files.
    """
    os.makedirs(sample_path, exist_ok=True)
    for c_name in queues.keys():
        dest_base = os.path.join(sample_path, "{}.{}".format(c_name, CorpusName.SAMPLE))
        with open(dest_base + ".src", "w", encoding="utf-8") as f_src, open(
            dest_base + ".tgt", "w", encoding="utf-8"
        ) as f_tgt:
            while True:
                _next = False
                for q in queues[c_name]:
                    item = q.get()
                    if item == "blank":
                        continue
                    if item == "break":
                        _next = True
                        break
                    _, src_line, tgt_line = item
                    f_src.write(src_line + "\n")
                    f_tgt.write(tgt_line + "\n")
                if _next:
                    break


def build_sub_vocab(corpora, transforms, config, n_sample, stride, offset):
    """Build vocab on (strided) subpart of the data."""
    sub_counter_src = Counter()
    sub_counter_tgt = Counter()
    datasets_iterables = build_corpora_iters(
        corpora,
        transforms,
        config.data,
        skip_empty_level=config.skip_empty_level,
        stride=stride,
        offset=offset,
    )
    for c_name, c_iter in datasets_iterables.items():
        for i, item in enumerate(c_iter):
            maybe_example = transform_bucket(CorpusTask.TRAIN, [item])
            if maybe_example is not None:
                maybe_example = maybe_example[0]
            else:
                if config.dump_samples:
                    build_sub_vocab.queues[c_name][offset].put("blank")
                continue
            src_line, tgt_line = (
                maybe_example["src"]["src"],
                maybe_example["tgt"]["tgt"],
            )
            sub_counter_src.update(src_line.split(" "))
            sub_counter_tgt.update(tgt_line.split(" "))

            if config.dump_samples:
                src_pretty_line = src_line
                build_sub_vocab.queues[c_name][offset].put((i, src_pretty_line, tgt_line))
            if n_sample > 0 and ((i + 1) * stride + offset) >= n_sample:
                if config.dump_samples:
                    build_sub_vocab.queues[c_name][offset].put("break")
                break
        if config.dump_samples:
            build_sub_vocab.queues[c_name][offset].put("break")
    return sub_counter_src, sub_counter_tgt


def init_pool(queues):
    """Add the queues as attribute of the pooled function."""
    build_sub_vocab.queues = queues


def build_vocab(config, transforms, n_sample=3):
    """Build vocabulary from data."""

    if n_sample == -1:
        logger.info(f"n_sample={n_sample}: Build vocab on full datasets.")
    elif n_sample > 0:
        logger.info(f"Build vocab on {n_sample} transformed examples/corpus.")
    else:
        raise ValueError(f"n_sample should > 0 or == -1, get {n_sample}.")

    if config.dump_samples:
        logger.info("The samples on which the vocab is built will be " "dumped to disk. It may slow down the process.")
    corpora = get_corpora(config, task=CorpusTask.TRAIN)
    counter_src = Counter()
    counter_tgt = Counter()

    queues = {
        c_name: [mp.Queue(config.vocab_sample_queue_size) for i in range(config.num_threads)]
        for c_name in corpora.keys()
    }
    if config.dump_samples:
        sample_path = os.path.join(config.save_data, CorpusName.SAMPLE)
        write_process = mp.Process(target=write_files_from_queues, args=(sample_path, queues), daemon=True)
        write_process.start()
    with mp.Pool(config.num_threads, init_pool, [queues]) as p:
        func = partial(build_sub_vocab, corpora, transforms, config, n_sample, config.num_threads)
        for sub_counter_src, sub_counter_tgt in p.imap(func, range(0, config.num_threads)):
            counter_src.update(sub_counter_src)
            counter_tgt.update(sub_counter_tgt)
    if config.dump_samples:
        write_process.join()
    return counter_src, counter_tgt


def ingest_tokens(config, transforms, n_sample, learner, stride, offset):
    def _mp_ingest(data):
        func = partial(transform_bucket, CorpusName.TRAIN)
        chunk = len(data) // config.num_threads
        with mp.Pool(config.num_threads) as pool:
            buckets = pool.map(
                func,
                [data[i * chunk : (i + 1) * chunk] for i in range(0, config.num_threads)],
            )
        for bucket in buckets:
            for ex in bucket:
                if ex is not None:
                    src_line, tgt_line = (ex["src"]["src"], ex["tgt"]["tgt"])
                    learner.ingest(src_line)
                    learner.ingest(tgt_line)

    corpora = get_corpora(config, task=CorpusTask.TRAIN)
    datasets_iterables = build_corpora_iters(
        corpora,
        transforms,
        config.data,
        skip_empty_level=config.skip_empty_level,
        stride=stride,
        offset=offset,
    )
    to_ingest = []
    for c_name, c_iter in datasets_iterables.items():
        for i, item in enumerate(c_iter):
            if n_sample >= 0 and i >= n_sample:
                break
            if len(to_ingest) >= MAXBUCKETSIZE:
                _mp_ingest(to_ingest)
                to_ingest = []
            to_ingest.append(item)
        _mp_ingest(to_ingest)


def make_learner(tokenization_type, symbols):
    if tokenization_type == "bpe":
        # BPE training
        learner = pyonmttok.BPELearner(tokenizer=None, symbols=symbols)
    elif tokenization_type == "sentencepiece":
        # SentencePiece training
        learner = pyonmttok.SentencePieceLearner(vocab_size=symbols, character_coverage=0.98)
    return learner


def build_vocab_main(config):
    """Apply transforms to samples of specified data and build vocab from it.

    Transforms that need vocab will be disabled in this.
    Built vocab is saved in plain text format as following and can be pass as
    `-src_vocab` (and `-tgt_vocab`) when training:
    ```
    <tok_0>\t<count_0>
    <tok_1>\t<count_1>
    ```
    """

    config._validate_data_config(build_vocab_only=True)
    assert config.n_sample == -1 or config.n_sample > 1, f"Illegal argument n_sample={config.n_sample}."

    logger = init_logger()
    set_random_seed(config.seed, False)
    transforms_cls = get_transforms_cls(config._all_transform)

    if config.learn_subwords:
        # if {"bpe", "sentencepiece", "onmt_tokenize"}.intersect(config.transforms):
        _config = config.transforms_configs.onmt_tokenize
        logger.info(f"Ingesting {_config.src_subword_type} model from corpus")
        learner = make_learner(_config.src_subword_type, config.learn_subwords_size)
        if _config.src_subword_model is not None:
            tok_path = _config.src_subword_model
        else:
            data_dir = os.path.split(config.save_data)[0]
            if not os.path.exists(data_dir):
                os.makedirs(data_dir)
            tok_path = os.path.join(data_dir, f"{_config.src_subword_type}.model")
        save_config = copy.deepcopy(config)
        _config.src_subword_type = "none"
        _config.tgt_subword_type = "none"
        _config.src_eoletok_kwargs["joiner_annotate"] = False
        _config.tgt_eoletok_kwargs["joiner_annotate"] = False
        transforms = make_transforms(config, transforms_cls, None)
        ingest_tokens(config, transforms, config.n_sample, learner, 1, 0)
        logger.info(f"Learning {tok_path} model, patience")
        learner.learn(tok_path)
        config = save_config

    transforms = make_transforms(config, transforms_cls, None)

    logger.info(f"Counter vocab from {config.n_sample} samples.")
    src_counter, tgt_counter = build_vocab(config, transforms, n_sample=config.n_sample)

    logger.info(f"Counters src: {len(src_counter)}")
    logger.info(f"Counters tgt: {len(tgt_counter)}")

    def save_counter(counter, save_path):
        check_path(save_path, exist_ok=config.overwrite, log=logger.warning)
        with open(save_path, "wb") as fo:
            for tok, count in counter.most_common():
                fo.write(f"{tok}\t{str(count)}\n".encode("utf-8"))

    if config.share_vocab:
        src_counter += tgt_counter
        tgt_counter = src_counter
        logger.info(f"Counters after share:{len(src_counter)}")
        save_counter(src_counter, config.src_vocab)
    else:
        save_counter(src_counter, config.src_vocab)
        save_counter(tgt_counter, config.tgt_vocab)


@register_bin(name="build_vocab")
class BuildVocab(RunBin):
    config_class = BuildVocabConfig
    require_config = True

    @classmethod
    def run(cls, args):
        config = cls.build_config(args)
        build_vocab_main(config)


def _get_parser():
    # Patch to allow autodoc with sphinx-argparse
    parser = ArgumentParser()
    BuildVocab.add_args(parser)
    return parser



"""Base class for encoders and generic multi encoders."""

import torch.nn as nn


class EncoderBase(nn.Module):
    """
    Base encoder class. Specifies the interface used by different encoder types
    and required by :class:
    `eole.Models.EncoderDecoderModel`
    `eole.Models.EncoderModel`

    """

    @classmethod
    def from_config(cls, encoder_config, running_config=None):
        raise NotImplementedError

    def forward(self, emb, **kwargs):
        """
        Args:
            emb (FloatTensor):
               embeddings ``(batch, src_len, dim)``
            **kwargs
               pad_mask (BoolTensor):
                  pad_mask ``(batch, maxlen)`` False when value, True when pad
        Returns:
            (FloatTensor, FloatTensor, FloatTensor):

            * enc_out (encoder output used for attention),
              ``(batch, src_len, hidden_size)``
              for bidirectional rnn last dimension is 2x hidden_size
            * enc_final_hs: encoder final hidden state
              ``(num_layers x dir, batch, hidden_size)``
              In the case of LSTM this is a tuple.
        """

        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Most code taken from: https://github.com/alvations/sacremoses
# Which in turn is based on the Moses punctuation normalizer.
# https://github.com/moses-smt/mosesdecoder/blob/master/scripts/
# tokenizer/normalize-punctuation.perl


from eole.utils.logging import logger
import re
import regex

from itertools import chain

from eole.transforms import register_transform
from .transform import Transform, TransformConfig
from pydantic import Field


class NormalizeConfig(TransformConfig):
    src_lang: str | None = Field(default="", description="Source language code")
    tgt_lang: str | None = Field(default="", description="Target language code")
    penn: bool | None = Field(default=True, description="Penn substitution")
    norm_quote_commas: bool | None = Field(default=True, description="Normalize quotations and commas")
    norm_numbers: bool | None = Field(default=True, description="Normalize numbers")
    pre_replace_unicode_punct: bool | None = Field(default=False, description="Replace unicode punct")
    post_remove_control_chars: bool | None = Field(default=False, description="Remove control chars")


class MosesPunctNormalizer:
    """
    This is a Python port of the Moses punctuation normalizer
    """

    EXTRA_WHITESPACE = [
        (r"\r", r""),
        (r"\(", r" ("),
        (r"\)", r") "),
        (r" +", r" "),
        (r"\) ([.!:?;,])", r")\g<1>"),
        (r"\( ", r"("),
        (r" \)", r")"),
        (r"(\d) %", r"\g<1>%"),
        (r" :", r":"),
        (r" ;", r";"),
    ]

    NORMALIZE_UNICODE_IF_NOT_PENN = [(r"`", r"'"), (r"''", r' " ')]

    NORMALIZE_UNICODE = [
        ("„", r'"'),
        ("“", r'"'),
        ("”", r'"'),
        ("–", r"-"),
        ("—", r" - "),
        (r" +", r" "),
        ("´", r"'"),
        ("([a-zA-Z])‘([a-zA-Z])", r"\g<1>'\g<2>"),
        ("([a-zA-Z])’([a-zA-Z])", r"\g<1>'\g<2>"),
        ("‘", r"'"),
        ("‚", r"'"),
        ("’", r"'"),
        (r"''", r'"'),
        ("´´", r'"'),
        ("…", r"..."),
    ]

    FRENCH_QUOTES = [
        ("\u00A0«\u00A0", r'"'),
        ("«\u00A0", r'"'),
        ("«", r'"'),
        ("\u00A0»\u00A0", r'"'),
        ("\u00A0»", r'"'),
        ("»", r'"'),
    ]

    HANDLE_PSEUDO_SPACES = [
        ("\u00A0%", r"%"),
        ("nº\u00A0", "nº "),
        ("\u00A0:", r":"),
        ("\u00A0ºC", " ºC"),
        ("\u00A0cm", r" cm"),
        ("\u00A0\\?", "?"),
        ("\u00A0\\!", "!"),
        ("\u00A0;", r";"),
        (",\u00A0", r", "),
        (r" +", r" "),
    ]

    EN_QUOTATION_FOLLOWED_BY_COMMA = [(r'"([,.]+)', r'\g<1>"')]

    DE_ES_FR_QUOTATION_FOLLOWED_BY_COMMA = [
        (r',"', r'",'),
        (r'(\.+)"(\s*[^<])', r'"\g<1>\g<2>'),
        # don't fix period at end of sentence
    ]

    DE_ES_CZ_CS_FR = [
        ("(\\d)\u00A0(\\d)", r"\g<1>,\g<2>"),
    ]

    OTHER = [
        ("(\\d)\u00A0(\\d)", r"\g<1>.\g<2>"),
    ]

    # Regex substitutions from replace-unicode-punctuation.perl
    # https://github.com/moses-smt/mosesdecoder/blob/master/
    # scripts/tokenizer/replace-unicode-punctuation.perl
    REPLACE_UNICODE_PUNCTUATION = [
        ("，", ","),
        (r"。\s*", ". "),
        ("、", ","),
        ("”", '"'),
        ("“", '"'),
        ("∶", ":"),
        ("：", ":"),
        ("？", "?"),
        ("《", '"'),
        ("》", '"'),
        ("）", ")"),
        ("！", "!"),
        ("（", "("),
        ("；", ";"),
        ("」", '"'),
        ("「", '"'),
        ("０", "0"),
        ("１", "1"),
        ("２", "2"),
        ("３", "3"),
        ("４", "4"),
        ("５", "5"),
        ("６", "6"),
        ("７", "7"),
        ("８", "8"),
        ("９", "9"),
        (r"．\s*", ". "),
        ("～", "~"),
        ("’", "'"),
        ("…", "..."),
        ("━", "-"),
        ("〈", "<"),
        ("〉", ">"),
        ("【", "["),
        ("】", "]"),
        ("％", "%"),
    ]

    def __init__(self):
        """ """

    def normalize(
        self,
        text,
        lang="en",
        penn=True,
        norm_quote_commas=True,
        norm_numbers=True,
        pre_replace_unicode_punct=False,
        post_remove_control_chars=False,
    ):
        """
        Returns a string with normalized punctuation.
        :param language: The two-letter language code.
        :type lang: str
        :param penn: Normalize Penn Treebank style quotations.
        :type penn: bool
        :param norm_quote_commas: Normalize quotations and commas
        :type norm_quote_commas: bool
        :param norm_numbers: Normalize numbers
        :type norm_numbers: bool
        """
        self.substitutions = [
            self.EXTRA_WHITESPACE,
            self.NORMALIZE_UNICODE,
            self.FRENCH_QUOTES,
            self.HANDLE_PSEUDO_SPACES,
        ]

        if penn:
            # Adds the penn substitutions after extra_whitespace regexes.
            self.substitutions.insert(1, self.NORMALIZE_UNICODE_IF_NOT_PENN)

        if norm_quote_commas:
            if lang == "en":
                self.substitutions.append(self.EN_QUOTATION_FOLLOWED_BY_COMMA)
            elif lang in ["de", "es", "fr"]:
                self.substitutions.append(self.DE_ES_FR_QUOTATION_FOLLOWED_BY_COMMA)

        if norm_numbers:
            if lang in ["de", "es", "cz", "cs", "fr"]:
                self.substitutions.append(self.DE_ES_CZ_CS_FR)
            else:
                self.substitutions.append(self.OTHER)

        self.substitutions = list(chain(*self.substitutions))

        # Optionally, replace unicode puncts BEFORE normalization.
        if pre_replace_unicode_punct:
            text = self.replace_unicode_punct(text)

        # Actual normalization.
        for regexp, substitution in self.substitutions:
            text = re.sub(regexp, substitution, str(text))

        # Optionally, replace unicode puncts BEFORE normalization.
        if post_remove_control_chars:
            text = self.remove_control_chars(text)

        return text.strip()

    def replace_unicode_punct(self, text):
        for regexp, substitution in self.REPLACE_UNICODE_PUNCTUATION:
            text = re.sub(regexp, substitution, str(text))
        return text

    def remove_control_chars(self, text):
        return regex.sub(r"\p{C}", "", text)


@register_transform(name="normalize")
class NormalizeTransform(Transform):
    """
    Normalize source and target based on Moses script.
    """

    config_model = NormalizeConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.src_lang = self.config.src_lang
        self.tgt_lang = self.config.tgt_lang
        self.penn = self.config.penn
        self.norm_quote_commas = self.config.norm_quote_commas
        self.norm_numbers = self.config.norm_numbers
        self.pre_replace_unicode_punct = self.config.pre_replace_unicode_punct
        self.post_remove_control_chars = self.config.post_remove_control_chars

    @staticmethod
    def _get_param(corpus, param, def_val):
        """Get opt string of a `corpus`."""
        if "normalize" in getattr(corpus, "transforms", []):
            value = getattr(corpus, param, def_val)
            normalize = value
        else:
            normalize = None
        return normalize

    @classmethod
    def get_config_dict(cls, config, param, def_val):
        """Get normalize settings correspond to corpus in `opts`."""
        normalize_dict = {}
        # normalize dict src/tgt for each dataset
        if hasattr(config, "data"):
            for c_name, corpus in config.data.items():
                normalize = cls._get_param(corpus, param, def_val)
                if normalize is not None:
                    logger.debug(f"Get {param} for {c_name}: {normalize}")
                    normalize_dict[c_name] = normalize

        return normalize_dict

    def warm_up(self, vocabs=None):
        """Set options for each dataset."""
        super().warm_up(None)
        self.src_lang_dict = self.get_config_dict(self.full_config, "src_lang", "")
        self.tgt_lang_dict = self.get_config_dict(self.full_config, "tgt_lang", "")
        self.penn_dict = self.get_config_dict(self.full_config, "penn", True)
        self.norm_quote_commas_dict = self.get_config_dict(self.full_config, "norm_quote_commas", True)
        self.norm_numbers_dict = self.get_config_dict(self.full_config, "norm_numbers", True)
        self.pre_dict = self.get_config_dict(self.full_config, "pre_replace_unicode_punct", False)
        self.post_dict = self.get_config_dict(self.full_config, "post_remove_control_chars", False)
        self.src_lang_dict["infer"] = self.src_lang
        self.tgt_lang_dict["infer"] = self.tgt_lang
        self.penn_dict["infer"] = self.penn
        self.norm_quote_commas_dict["infer"] = self.norm_quote_commas
        self.norm_numbers_dict["infer"] = self.norm_numbers
        self.pre_dict["infer"] = self.pre_replace_unicode_punct
        self.post_dict["infer"] = self.post_remove_control_chars

        self.mpn = MosesPunctNormalizer()

    def apply(self, example, is_train=False, stats=None, **kwargs):
        """Normalize source and target examples."""
        corpus_name = kwargs.get("corpus_name", None)
        if corpus_name is None:
            raise ValueError("corpus_name is required.")

        src_str = self.mpn.normalize(
            " ".join(example["src"]),
            self.src_lang_dict[corpus_name],
            self.penn_dict[corpus_name],
            self.norm_quote_commas_dict[corpus_name],
            self.norm_numbers_dict[corpus_name],
            self.pre_dict[corpus_name],
            self.post_dict[corpus_name],
        )
        example["src"] = src_str.split(" ")

        if example["tgt"] is not None:
            tgt_str = self.mpn.normalize(
                " ".join(example["tgt"]),
                self.tgt_lang_dict[corpus_name],
                self.penn_dict[corpus_name],
                self.norm_quote_commas_dict[corpus_name],
                self.norm_numbers_dict[corpus_name],
                self.pre_dict[corpus_name],
                self.post_dict[corpus_name],
            )
            example["tgt"] = tgt_str.split(" ")

        return example



"""Transforms relate to tokenization/subword."""

import re
from eole.utils.logging import logger
from eole.transforms import register_transform
from .transform import Transform, ObservableStats, TransformConfig
from eole.constants import DefaultTokens
from typing import Literal, List, Tuple
from pydantic import model_validator, Field


class BaseTokenizerConfig(TransformConfig):
    src_subword_model: str | None = Field(default=None, description="Path of subword model for src (or shared).")
    tgt_subword_model: str | None = Field(default=None, description="Path of subword model for tgt.")

    src_subword_nbest: int | None = Field(
        default=1,
        description="Number of candidates in subword regularization. "
        "Valid for unigram sampling, invalid for BPE-dropout. (source side)",
    )
    tgt_subword_nbest: int | None = Field(
        default=1,
        description="Number of candidates in subword regularization. "
        "Valid for unigram sampling, invalid for BPE-dropout. (target side)",
    )
    src_subword_alpha: float | None = Field(
        default=0,
        description="Smoothing parameter for sentencepiece unigram sampling, "
        "and dropout probability for BPE-dropout. (source side)",
    )
    tgt_subword_alpha: float | None = Field(
        default=0,
        description="Smoothing parameter for sentencepiece unigram sampling, "
        "and dropout probability for BPE-dropout. (target side)",
    )

    src_subword_vocab: str | None = Field(
        default="",
        description="Path to the vocabulary file for src subword. " "Format: <word>\\t<count> per line.",
    )
    tgt_subword_vocab: str | None = Field(
        default="",
        description="Path to the vocabulary file for tgt subword. " "Format: <word>\\t<count> per line.",
    )
    src_vocab_threshold: int | None = Field(
        default=0,
        description="Only produce src subword in src_subword_vocab " "with frequency >= src_vocab_threshold.",
    )
    tgt_vocab_threshold: int | None = Field(
        default=0,
        description="Only produce tgt subword in tgt_subword_vocab " "with frequency >= tgt_vocab_threshold.",
    )

    @model_validator(mode="after")
    def check_values(self):
        # TODO: use numeric constraints on Field instead?
        # https://docs.pydantic.dev/latest/concepts/fields/#numeric-constraints
        assert 0 <= self.src_subword_alpha <= 1, "src_subword_alpha should be in the range [0, 1]"
        assert 0 <= self.tgt_subword_alpha <= 1, "tgt_subword_alpha should be in the range [0, 1]"
        return self


class ONMTTokenizerConfig(BaseTokenizerConfig):
    src_subword_type: Literal["none", "sentencepiece", "bpe"] | None = Field(
        default="none",
        description="Type of subword model for src (or shared) in pyonmttok.",
    )
    tgt_subword_type: Literal["none", "sentencepiece", "bpe"] | None = Field(
        default="none", description="Type of subword model for tgt in pyonmttok."
    )
    src_onmttok_kwargs: dict | None = Field(
        default={"mode": "none"},
        description="Other pyonmttok options for src in dict string, " "except subword related options listed earlier.",
    )
    tgt_onmttok_kwargs: dict | None = Field(
        default={"mode": "none"},
        description="Other pyonmttok options for tgt in dict string, " "except subword related options listed earlier.",
    )
    gpt2_pretok: bool | None = Field(default=False, description="Preprocess sentence with byte-level mapping.")
    mapped_tokens: List[Tuple[str, str]] | None = Field(
        default=None,
        description="Mapped tokens for placeholders preservation",
    )

    @model_validator(mode="after")
    def check_values(self):
        # TODO: this can probably be enhanced
        if isinstance(self.src_onmttok_kwargs, str):
            self.src_onmttok_kwargs = eval(self.src_onmttok_kwargs)
        if isinstance(self.tgt_onmttok_kwargs, str):
            self.tgt_onmttok_kwargs = eval(self.tgt_onmttok_kwargs)
        if not isinstance(self.src_onmttok_kwargs, dict):
            raise ValueError("-src_onmttok_kwargs isn't a dict valid string.")
        if not isinstance(self.tgt_onmttok_kwargs, dict):
            raise ValueError("-tgt_onmttok_kwargs isn't a dict valid string.")
        return self


class TokenizerTransform(Transform):
    """Tokenizer transform abstract class."""

    config_model = BaseTokenizerConfig

    def __init__(self, config):
        """Initialize necessary options for Tokenizer."""
        super().__init__(config)

    def _parse_config(self):
        self.share_vocab = self.full_config.share_vocab
        self.src_subword_model = self.config.src_subword_model
        self.tgt_subword_model = self.config.tgt_subword_model
        self.src_subword_nbest = self.config.src_subword_nbest
        self.tgt_subword_nbest = self.config.tgt_subword_nbest
        self.src_subword_alpha = self.config.src_subword_alpha
        self.tgt_subword_alpha = self.config.tgt_subword_alpha
        self.src_subword_vocab = self.config.src_subword_vocab
        self.tgt_subword_vocab = self.config.tgt_subword_vocab
        self.src_vocab_threshold = self.config.src_vocab_threshold
        self.tgt_vocab_threshold = self.config.tgt_vocab_threshold
        # define which attributes are artifacts to be saved
        self.artifacts = [
            "src_subword_model",
            "tgt_subword_model",
            "src_subword_vocab",
            "tgt_subword_vocab",
        ]

    def _repr_args(self):
        """Return str represent key arguments for TokenizerTransform."""
        kwargs = {
            "share_vocab": self.share_vocab,
            "src_subword_model": self.src_subword_model,
            "tgt_subword_model": self.tgt_subword_model,
            "src_subword_alpha": self.src_subword_alpha,
            "tgt_subword_alpha": self.tgt_subword_alpha,
            "src_subword_vocab": self.src_subword_vocab,
            "tgt_subword_vocab": self.tgt_subword_vocab,
            "src_vocab_threshold": self.src_vocab_threshold,
            "tgt_vocab_threshold": self.tgt_vocab_threshold,
        }
        return ", ".join([f"{kw}={arg}" for kw, arg in kwargs.items()])

    def warm_up(self, vocabs=None):
        super().warm_up(None)
        if vocabs is not None:
            self.pad_token = vocabs["specials"].get("pad_token", DefaultTokens.PAD)
            self.eos_token = vocabs["specials"].get("eos_token", DefaultTokens.EOS)
        else:
            self.pad_token = DefaultTokens.PAD
            self.eos_token = DefaultTokens.EOS

    def tokenize_string(self, string, side="src", is_train=False):
        raise NotImplementedError

    def _tokenize(self, tokens, side="src", is_train=False):
        """Tokenize a list of words."""
        # This method embeds a custom logic to correctly handle certain placeholders
        # in case the tokenizer doesn't preserve them.
        sentence = " ".join(tokens)
        # We need to split sentence on EOS and mapped_tokens
        # to make sure sentencepiece add a joiner just after special tokens
        if getattr(self, "mapped_tokens", None) is not None:
            delim_list = [mapped_toks[0] for mapped_toks in self.mapped_tokens] + [self.eos_token]
        else:
            delim_list = [self.eos_token]
        pattern = f"({'|'.join(map(re.escape, delim_list))})"
        sent_list = re.split(pattern, sentence)
        # remove empty elements
        sent_list = [item for item in sent_list if item]

        # Tokenize each sentence separately.
        segmented = []
        for _sentence in sent_list:
            if _sentence in delim_list:
                segmented.append(_sentence)
            else:
                _sentence_tokens = self.tokenize_string(_sentence, side, is_train)
                if _sentence[-1] == " ":
                    trailingspace = self.tokenize_string(" a", side, is_train)[0][:-1]
                    _sentence_tokens += [trailingspace]
                segmented.extend(_sentence_tokens)
        return segmented

    def apply(self, example, is_train=False, stats=None, **kwargs):
        """Apply subword-based tokenenization to src & tgt."""
        src_out = self._tokenize(example["src"], "src", is_train)
        if example["tgt"] is not None:
            tgt_out = self._tokenize(example["tgt"], "tgt", is_train)
            if stats is not None:
                n_words = len(example["src"]) + len(example["tgt"])
                n_subwords = len(src_out) + len(tgt_out)
                stats.update(SubwordStats(n_subwords, n_words))
        else:
            tgt_out = None
            if stats is not None:
                n_words = len(example["src"])
                n_subwords = len(src_out)
                stats.update(SubwordStats(n_subwords, n_words))
        example["src"], example["tgt"] = src_out, tgt_out
        return example


class SubwordStats(ObservableStats):
    """Runing statistics for counting tokens before/after subword transform."""

    __slots__ = ["subwords", "words"]

    def __init__(self, subwords: int, words: int):
        self.subwords = subwords
        self.words = words

    def update(self, other: "SubwordStats"):
        self.subwords += other.subwords
        self.words += other.words

    def __str__(self) -> str:
        return "{}: {} -> {} tokens".format(self.name(), self.words, self.subwords)


@register_transform(name="sentencepiece")
class SentencePieceTransform(TokenizerTransform):
    """SentencePiece subword transform class."""

    def __init__(self, config):
        """Initialize necessary options for sentencepiece."""
        super().__init__(config)
        self.mapped_tokens = []

    def _set_seed(self, seed):
        """set seed to ensure reproducibility."""
        import sentencepiece as spm

        spm.set_random_generator_seed(seed)

    def warm_up(self, vocabs=None):
        """Load subword models."""
        super().warm_up(None)
        import sentencepiece as spm

        load_src_model = spm.SentencePieceProcessor()
        load_src_model.Load(self.src_subword_model)
        _diff_vocab = (
            self.src_subword_vocab != self.tgt_subword_vocab or self.src_vocab_threshold != self.tgt_vocab_threshold
        )
        if self.src_subword_vocab != "" and self.src_vocab_threshold > 0:
            load_src_model.LoadVocabulary(self.src_subword_vocab, self.src_vocab_threshold)
        if self.share_vocab and not _diff_vocab:
            self.load_models = {"src": load_src_model, "tgt": load_src_model}
        else:
            load_tgt_model = spm.SentencePieceProcessor()
            load_tgt_model.Load(self.tgt_subword_model)
            if self.tgt_subword_vocab != "" and self.tgt_vocab_threshold > 0:
                load_tgt_model.LoadVocabulary(self.tgt_subword_vocab, self.tgt_vocab_threshold)
            self.load_models = {"src": load_src_model, "tgt": load_tgt_model}

    def tokenize_string(self, string, side="src", is_train=False):
        """Apply subword sampling or deterministic subwording"""
        sp_model = self.load_models[side]
        nbest_size = self.tgt_subword_nbest if side == "tgt" else self.src_subword_nbest
        string = string.replace(DefaultTokens.SEP, "\n")
        if is_train is False or nbest_size in [0, 1]:
            # derterministic subwording
            tokens = sp_model.encode(string, out_type=str)
        else:
            # subword sampling when nbest_size > 1 or -1
            # alpha should be 0.0 < alpha < 1.0
            alpha = self.tgt_subword_alpha if side == "tgt" else self.src_subword_alpha
            tokens = sp_model.encode(
                string,
                out_type=str,
                enable_sampling=True,
                alpha=alpha,
                nbest_size=nbest_size,
            )
        return tokens

    def _detokenize(self, tokens, side="src"):
        """Apply SentencePiece Detokenizer"""
        sp_model = self.load_models[side]
        return sp_model.DecodePieces(tokens).replace("\n", DefaultTokens.SEP)

    def apply_reverse(self, predicted):
        """Apply SentencePiece Detokenizer."""
        if isinstance(predicted, list):
            return self._detokenize(predicted, "tgt")
        else:
            return self._detokenize(predicted.split(" "), "tgt")

    def _repr_args(self):
        """Return str represent key arguments for class."""
        kwargs_str = super()._repr_args()
        additional_str = "src_subword_nbest={}, tgt_subword_nbest={}".format(
            self.src_subword_nbest, self.tgt_subword_nbest
        )
        return kwargs_str + ", " + additional_str


@register_transform(name="bpe")
class BPETransform(TokenizerTransform):
    """subword_nmt: official BPE subword transform class."""

    def __init__(self, config):
        """Initialize necessary options for subword_nmt."""
        super().__init__(config)

    def _parse_config(self):
        super()._parse_config()
        self.dropout = {"src": self.src_subword_alpha, "tgt": self.tgt_subword_alpha}
        self.mapped_tokens = []

    def _set_seed(self, seed):
        """set seed to ensure reproducibility."""
        import random

        random.seed(seed)

    def warm_up(self, vocabs=None):
        """Load subword models."""
        super().warm_up(None)
        from subword_nmt.apply_bpe import BPE, read_vocabulary

        # Load vocabulary file if provided and set threshold
        src_vocabulary, tgt_vocabulary = None, None
        if self.src_subword_vocab != "" and self.src_vocab_threshold > 0:
            with open(self.src_subword_vocab, encoding="utf-8") as _sv:
                src_vocabulary = read_vocabulary(_sv, self.src_vocab_threshold)
        if self.tgt_subword_vocab != "" and self.tgt_vocab_threshold > 0:
            with open(self.tgt_subword_vocab, encoding="utf-8") as _tv:
                tgt_vocabulary = read_vocabulary(_tv, self.tgt_vocab_threshold)
        # Load Subword Model
        with open(self.src_subword_model, encoding="utf-8") as src_codes:
            load_src_model = BPE(codes=src_codes, vocab=src_vocabulary)
        if self.share_vocab and (src_vocabulary == tgt_vocabulary):
            self.load_models = {"src": load_src_model, "tgt": load_src_model}
        else:
            with open(self.tgt_subword_model, encoding="utf-8") as tgt_codes:
                load_tgt_model = BPE(codes=tgt_codes, vocab=tgt_vocabulary)
            self.load_models = {"src": load_src_model, "tgt": load_tgt_model}

    def tokenize_string(self, string, side="src", is_train=False):
        """Do bpe subword tokenize."""
        tokens = string.split(" ")
        bpe_model = self.load_models[side]
        dropout = self.dropout[side] if is_train else 0.0
        segmented = bpe_model.segment_tokens(tokens, dropout=dropout)
        return segmented

    def _detokenize(self, tokens, side="src", is_train=False):
        """ "Apply bpe subword detokenizer"""
        detokenized = re.sub(r"(@@ )|(@@ ?$)", r"", " ".join(tokens))
        return detokenized

    def apply_reverse(self, predicted):
        """Apply bpe subword detokenizer"""
        if isinstance(predicted, list):
            return self._detokenize(predicted, "tgt")
        else:
            return self._detokenize(predicted.split(" "), "tgt")


@register_transform(name="onmt_tokenize")
class ONMTTokenizerTransform(TokenizerTransform):
    """OpenNMT Tokenizer transform class."""

    # config model as class attribute
    config_model = ONMTTokenizerConfig

    def __init__(self, config):
        """Initialize necessary options for OpenNMT Tokenizer."""
        super().__init__(config)

    def _set_seed(self, seed):
        """set seed to ensure reproducibility."""
        import pyonmttok

        pyonmttok.set_random_seed(seed)

    def _parse_config(self):
        super()._parse_config()
        self.src_subword_type = self.config.src_subword_type
        self.tgt_subword_type = self.config.tgt_subword_type
        logger.debug("Parsed pyonmttok kwargs for src: {}".format(self.config.src_onmttok_kwargs))
        logger.debug("Parsed pyonmttok kwargs for tgt: {}".format(self.config.tgt_onmttok_kwargs))
        self.src_other_kwargs = self.config.src_onmttok_kwargs
        self.tgt_other_kwargs = self.config.tgt_onmttok_kwargs
        self.gpt2_pretok = self.config.gpt2_pretok
        self.mapped_tokens = self.config.mapped_tokens
        self.preserve_placeholders = self.config.tgt_onmttok_kwargs.get("preserve_placeholders", False)

    @classmethod
    def get_specials(cls, config):
        config = getattr(config.transforms_configs, cls.name)
        src_specials, tgt_specials = [], []
        if config.src_onmttok_kwargs.get("case_markup", False):
            _case_specials = [
                "｟mrk_case_modifier_C｠",
                "｟mrk_begin_case_region_U｠",
                "｟mrk_end_case_region_U｠",
            ]
            for src_spec in _case_specials:
                src_specials.append(src_spec)
        if config.tgt_onmttok_kwargs.get("case_markup", False):
            _case_specials = [
                "｟mrk_case_modifier_C｠",
                "｟mrk_begin_case_region_U｠",
                "｟mrk_end_case_region_U｠",
            ]
            for tgt_spec in _case_specials:
                tgt_specials.append(tgt_spec)
        return (src_specials, tgt_specials)

    def _get_subword_kwargs(self, side="src"):
        """Return a dict containing kwargs relate to `side` subwords."""
        subword_type = self.tgt_subword_type if side == "tgt" else self.src_subword_type
        subword_model = self.tgt_subword_model if side == "tgt" else self.src_subword_model
        subword_nbest = self.tgt_subword_nbest if side == "tgt" else self.src_subword_nbest
        subword_alpha = self.tgt_subword_alpha if side == "tgt" else self.src_subword_alpha
        kwargs = dict()
        if subword_type == "bpe":
            kwargs["bpe_model_path"] = subword_model
            kwargs["bpe_dropout"] = subword_alpha
        elif subword_type == "sentencepiece":
            kwargs["sp_model_path"] = subword_model
            kwargs["sp_nbest_size"] = subword_nbest
            kwargs["sp_alpha"] = subword_alpha
        else:
            logger.debug("No subword method will be applied.")
        vocabulary_threshold = self.tgt_vocab_threshold if side == "tgt" else self.src_vocab_threshold
        vocabulary_path = self.tgt_subword_vocab if side == "tgt" else self.src_subword_vocab
        if vocabulary_threshold > 0 and vocabulary_path != "":
            kwargs["vocabulary_path"] = vocabulary_path
            kwargs["vocabulary_threshold"] = vocabulary_threshold
        return kwargs

    def warm_up(self, vocabs=None):
        """Initialize Tokenizer models."""
        super().warm_up(None)
        import pyonmttok

        src_subword_kwargs = self._get_subword_kwargs(side="src")
        src_tokenizer = pyonmttok.Tokenizer(**src_subword_kwargs, **self.src_other_kwargs)
        tgt_subword_kwargs = self._get_subword_kwargs(side="tgt")
        _diff_vocab = src_subword_kwargs.get("vocabulary_path", "") != tgt_subword_kwargs.get(
            "vocabulary_path", ""
        ) or src_subword_kwargs.get("vocabulary_threshold", 0) != tgt_subword_kwargs.get("vocabulary_threshold", 0)
        if self.share_vocab and not _diff_vocab:
            self.load_models = {"src": src_tokenizer, "tgt": src_tokenizer}
        else:
            tgt_subword_kwargs = self._get_subword_kwargs(side="tgt")
            tgt_tokenizer = pyonmttok.Tokenizer(**tgt_subword_kwargs, **self.tgt_other_kwargs)
            self.load_models = {"src": src_tokenizer, "tgt": tgt_tokenizer}
        if self.gpt2_pretok:
            """
            Returns list of utf-8 byte and a corresponding list of unicode
            strings. The reversible bpe codes work on unicode strings.
            code taken from openai/gpt2
            """
            bs = (
                list(range(ord("!"), ord("~") + 1))
                + list(range(ord("¡"), ord("¬") + 1))
                + list(range(ord("®"), ord("ÿ") + 1))
            )
            cs = bs[:]
            n = 0
            for b in range(2**8):
                if b not in bs:
                    bs.append(b)
                    cs.append(2**8 + n)
                    n += 1
            cs = [chr(n) for n in cs]
            self.maptable = dict(zip(bs, cs))
            self.revtable = {v: k for k, v in self.maptable.items()}

    def tokenize_string(self, sentence, side="src", is_train=False):
        tokenizer = self.load_models[side]

        if self.mapped_tokens is not None:
            for mapped_toks in self.mapped_tokens:
                sentence = sentence.replace(mapped_toks[0], mapped_toks[1])

        if self.gpt2_pretok:
            sentence = "".join(self.maptable[b] for b in sentence.replace(DefaultTokens.SEP, "\n").encode("utf-8"))
            sentence = sentence.replace("ï½Ł", "\uff5f").replace("ï½ł", "\uff60")
            segmented1 = tokenizer(sentence)
            segmented = []
            # ugly patch to make sure "\n\n" is split in two items
            for s in segmented1:
                if s == "ĊĊ":
                    segmented.extend(["Ċ", "Ċ"])
                else:
                    segmented.append(s)
        elif self.src_subword_type == "sentencepiece":  # and not self.preserve_placeholders
            sentence = sentence.replace(DefaultTokens.SEP, "\n")
            segmented = tokenizer(sentence)
        else:
            segmented = tokenizer(sentence)

        if self.mapped_tokens is not None:
            mapped_dict = {b: a for a, b in self.mapped_tokens}
            segmented = [mapped_dict.get(tok, tok) for tok in segmented]

        return segmented

    def _detokenize(self, tokens, side="src", is_train=False):
        """Do OpenNMT Tokenizer's detokenize."""

        def hex_to_char(match):
            # Extract all matched <0x..> sequences into a single hex string
            hex_sequence = match.group(0)  # Entire matched string
            # Remove `<0x` and `>` and any spaces
            hex_values = re.findall(r"<0x([0-9A-Fa-f]{2})>", hex_sequence)
            # Convert the hex values into a byte sequence
            byte_sequence = bytes.fromhex("".join(hex_values))
            # Decode the byte sequence into a UTF-8 string
            try:
                return byte_sequence.decode("utf-8")
            except UnicodeDecodeError:
                return "�"

        tokenizer = self.load_models[side]
        if self.gpt2_pretok:
            sentence = "".join(tokens)
            sentence = sentence.replace("｟", "ï½Ł").replace("｠", "ï½ł")
            detokenized = bytearray([self.revtable[c] for c in sentence]).decode("utf-8", errors="replace")
        else:
            detokenized = tokenizer.detokenize(tokens)
            # let's match one or more repetitions of the <0x..> pattern
            detokenized = re.sub(r"(?:<0x[0-9A-Fa-f]{2}>\s*)+", hex_to_char, detokenized)
        return detokenized.replace("\n", DefaultTokens.SEP)

    def apply_reverse(self, predicted):
        """Apply OpenNMT Tokenizer to src & tgt."""
        if isinstance(predicted, list):
            return self._detokenize(predicted, "tgt")
        else:
            return self._detokenize(predicted.split(" "), "tgt")

    def _repr_args(self):
        """Return str represent key arguments for class."""
        repr_str = "{}={}".format("share_vocab", self.share_vocab)
        repr_str += ", src_subword_kwargs={}".format(self._get_subword_kwargs(side="src"))
        repr_str += ", src_onmttok_kwargs={}".format(self.src_other_kwargs)
        repr_str += ", tgt_subword_kwargs={}".format(self._get_subword_kwargs(side="tgt"))
        repr_str += ", tgt_onmttok_kwargs={}".format(self.tgt_other_kwargs)
        return repr_str



from eole.constants import DefaultTokens
from eole.transforms import register_transform
from eole.utils.logging import logger
from .transform import Transform, TransformConfig
from pydantic import Field
from typing import List


class InsertMaskBeforePlaceholderConfig(TransformConfig):
    response_patterns: List[str] | None = Field(
        default=["Response : ｟newline｠"],
        description="Response pattern to locate the end of the prompt.",
    )


@register_transform(name="insert_mask_before_placeholder")
class InsertMaskBeforePlaceholdersTransform(Transform):
    """Add the `DefaultTokens.MASK_BEFORE` placeholder between
    the prompt and the response in an LM finetuning exemple.
    This is necessary to enable the 'zero-out prompt loss' mechanism.
    """

    config_model = InsertMaskBeforePlaceholderConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.response_patterns = self.config.response_patterns

    def apply(self, example, is_train=False, stats=None, **kwargs):
        _src = " ".join(example["src"])
        response = None
        for _pattern in self.response_patterns:
            if len(_src.split(_pattern)) == 2:
                prompt, response = _src.split(_pattern)
                response = DefaultTokens.MASK_BEFORE.join([_pattern, response])
        if response is not None:
            _src = "".join([prompt, response])
            example["src"] = _src.split(" ")
            example["tgt"] = _src.split(" ")
        else:
            logger.info("The mask_before could not be inserted")
            return example
        return example



#!/usr/bin/env python3
# coding: utf-8

# In order to use this tool, please install comet first
# https://github.com/Unbabel/COMET

# Let's say you have a source file with N sentences in SL - eg: source.sl
# and the corresponding references (N sentences) reference.tl
# Translate your file in TL with the -n_best nbest options nbest being
# then number of hypotheses and output the target to -output target.nbest.tl
# Then you need to duplicate source and reference sentences nbest times
# for this script.
# for instance using awk '{for(i=1; i<=n; i++) print}' n=5 reference.tl \
#                          > reference.5.tl
# same for source.
# This script can be run (for instance with nbest = 5) as follows:
# python oracle_bleu.py --nbest-src source.5.sl --nbest-hyp target.5.tl \
#  --nbest-ref reference.5.tl --nbest-order 5 --output target.maxbleu.tl
# It will search in all hyp the best comet score

import codecs
from eole.bin import BaseBin, register_bin


def chunks(lgth, n):
    """Yield successive n-sized chunks from lgth."""
    for i in range(0, len(lgth), n):
        yield lgth[i : i + n]


# when choosing a reference-less model no nbest-ref is required


@register_bin(name="oracle_comet")
class OracleComet(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument("--nbest-src", type=str, help="src repeated n times", required=True)
        parser.add_argument("--nbest-hyp", type=str, help="file with nbest to rerank", required=True)
        parser.add_argument("--nbest-ref", type=str, help="ref repeated n times", required=False)
        parser.add_argument("--nbest-order", type=int, help="nbest order", required=True)
        parser.add_argument("--output", type=str, help="output file", required=True)

    @classmethod
    def run(cls, args):
        from comet import download_model, load_from_checkpoint

        model_path = download_model("wmt21-comet-qe-mqm")
        model = load_from_checkpoint(model_path)

        with codecs.open(args.nbest_hyp, encoding="utf-8") as file:
            nbests = file.readlines()
            nbests = [item.strip() for item in nbests]
            nbests = chunks(nbests, args.nbest_order)

        if args.nbest_ref:
            with codecs.open(args.nbest_ref, encoding="utf-8") as file:
                nrefs = file.readlines()
                nrefs = [item.strip() for item in nrefs]
                nrefs = chunks(nrefs, args.nbest_order)

        with codecs.open(args.nbest_src, encoding="utf-8") as file:
            nsrcs = file.readlines()
            nsrcs = [item.strip() for item in nsrcs]
            nsrcs = chunks(nsrcs, args.nbest_order)

        with codecs.open(args.output, "w", encoding="utf-8") as output_file:
            best_indices = []
            # for nbest in nbests:
            if args.nbest_ref:
                for nsrc, nbest, nref in zip(nsrcs, nbests, nrefs):
                    data = []
                    for i in range(args.nbest_order):
                        data.append(
                            {
                                "src": nsrc[i],
                                "mt": nbest[i],
                                "ref": nref[i],
                            }
                        )
                        seg_scores, sys_score = model.predict(data, batch_size=args.nbest_order, gpus=1)
                        max_index = seg_scores.index(max(seg_scores))
                        output_file.write(data[max_index]["mt"] + "\n")
                        best_indices.append(max_index)
            else:
                for nsrc, nbest in zip(nsrcs, nbests):
                    data = []
                    for i in range(args.nbest_order):
                        data.append(
                            {
                                "src": nsrc[i],
                                "mt": nbest[i],
                            }
                        )
                        seg_scores, sys_score = model.predict(data, batch_size=args.nbest_order, gpus=1)
                        max_index = seg_scores.index(max(seg_scores))
                        output_file.write(data[max_index]["mt"] + "\n")
                        best_indices.append(max_index)

        for i in range(args.nbest_order):
            print(i, best_indices.count(i))



"""Module defining encoders."""

from eole.encoders.transformer import TransformerEncoder
from eole.encoders.rnn_encoder import RNNEncoder
from eole.encoders.cnn_encoder import CNNEncoder
from eole.encoders.mean_encoder import MeanEncoder
from eole.encoders.vision import VisionEncoder


str2enc = {
    "rnn": RNNEncoder,
    "brnn": RNNEncoder,
    "cnn": CNNEncoder,
    "transformer": TransformerEncoder,
    "mean": MeanEncoder,
    "vision": VisionEncoder,
}



import os
from typing import Dict, List, Literal
from pydantic import Field, field_validator, model_validator
from pydantic import create_model

from eole import constants
from eole.utils.logging import logger
from eole.transforms import AVAILABLE_TRANSFORMS
from eole.config.config import Config


class BaseVocabConfig(Config):
    src_vocab: str | None = Field(
        description="Path to src (or shared) vocabulary file. " "Format: one <word> or <word>\t<count> per line."
    )
    tgt_vocab: str | None = Field(
        default=None,
        description="Path to tgt vocabulary file. Format: one <word> or <word>\t<count> per line.",
    )
    share_vocab: bool = Field(default=False, description="Share source and target vocabulary.")
    decoder_start_token: str = Field(
        default=constants.DefaultTokens.BOS,
        description="Default decoder start token. For most models it is <s> = BOS. "
        "Some fairseq models require </s>.",
    )
    bos_token: str | None = Field(
        default=constants.DefaultTokens.BOS,
    )
    eos_token: str | None = Field(
        default=constants.DefaultTokens.EOS,
    )
    unk_token: str | None = Field(
        default=constants.DefaultTokens.UNK,
    )
    pad_token: str | None = Field(
        default=constants.DefaultTokens.PAD,
    )
    # pre trained embeddings stuff, might be put elsewhere
    both_embeddings: str | None = Field(
        default=None,
        description="Path to the embeddings file to use for both source and target tokens.",
    )
    src_embeddings: str | None = Field(
        default=None,
        description="Path to the embeddings file to use for source tokens.",
    )
    tgt_embeddings: str | None = Field(
        default=None,
        description="Path to the embeddings file to use for target tokens.",
    )
    embeddings_type: Literal["GloVe", "word2vec"] | None = Field(default=None, description="Type of embeddings file.")


class VocabConfig(BaseVocabConfig):
    src_vocab_size: int = Field(default=32758, description="Maximum size of the source vocabulary.")
    tgt_vocab_size: int = Field(default=32768, description="Maximum size of the target vocabulary.")
    vocab_size_multiple: int = Field(
        default=8,
        description="Make the vocabulary size a multiple of this value. " "(Adds dummy tokens if needed.)",
    )
    src_words_min_frequency: int = Field(default=0, description="Discard source words with lower frequency.")
    tgt_words_min_frequency: int = Field(default=0, description="Discard target words with lower frequency.")


class Dataset(Config):
    name: str | None = None  # we could put some auto name generation in validation (from path, uuid ,etc.) # noqa: E501
    weight: int | None = 1
    # streams: List[Stream] # maybe implemented later
    transforms: List[str] | None = None
    path_src: str | None = None
    path_tgt: str | None = None
    path_sco: str | None = None
    path_txt: str | None = None  # not sure we need this one with proper path_src handling
    path_align: str | None = None
    # optional stuff for some transforms
    # TODO: define a better mechanism to support such settings
    src_prefix: str | None = None
    tgt_prefix: str | None = None
    src_suffix: str | None = None
    tgt_suffix: str | None = None
    # normalize
    src_lang: str | None = None
    tgt_lang: str | None = None
    penn: bool | None = True
    norm_quote_commas: bool | None = True
    norm_numbers: bool | None = True
    pre_replace_unicode_punct: bool | None = False
    post_remove_control_chars: bool | None = False
    # clean
    src_eq_tgt: bool | None = True
    same_char: bool | None = True
    same_word: bool | None = True
    scripts_ok: List[str] | None = ["Latin", "Common"]
    scripts_nok: List[str] | None = []
    src_tgt_ratio: float | None = 2
    avg_tok_min: float | None = 3
    avg_tok_max: float | None = 20
    lang_id: List[str] | None = ["en", "fr"]


# add all opts from all transforms (like in eole.opts._add_transform_opt)
# maybe we want to restrict to the transforms actually used here to prevent overcrowding the config

# nested way of storing transform configs, for clarity
transforms_configs = {}
for name, transform_cls in AVAILABLE_TRANSFORMS.items():
    if hasattr(transform_cls, "config_model"):
        transforms_configs[name] = (
            transform_cls.config_model,
            transform_cls.config_model(),
        )


NestedAllTransformsConfig = create_model(
    "NestedAllTransformsConfig",
    __base__=(Config,),
    **transforms_configs,
)


class DataConfig(VocabConfig):  # , AllTransformsConfig):
    data: Dict[str, Dataset] | None = Field(
        description="All datasets and their specifications. " "See examples/*.yaml for further details."
    )
    transforms: List[str] = Field(
        default=[],
        description="Default transform pipeline to apply to data. "
        "Can be specified in each corpus of data to override.",
    )  # this should be validated against AVAILABLE_TRANSFORMS.keys() (Enum ?)
    transforms_configs: NestedAllTransformsConfig | None = Field(default_factory=NestedAllTransformsConfig)
    _all_transforms: List[str]
    # NameError: Fields must not use names with leading underscores
    # _all_transforms: List[str] = Field(
    #     default=None,
    #     description="Dummy field, programmatically deduced from "
    #                 "general and corpus-specific transforms."
    # )  # might be a cleaner way to do this
    skip_empty_level: Literal["silent", "warning", "error"] = Field(
        default="warning",
        description="Logging level when encoutering empty examples. "
        "(silent: silently ignore/skip empty examples, "
        "warning: warn when ignoring/skipping empty examples, "
        "error: raise an error and stop execution when any empty example)",
    )
    # patch to allow test_data_prepare to pass for now
    n_sample: int | None = 0  # mostly overridden in Vocab/Train configs
    save_data: str | None = Field(
        default=None,
        description="Output base path for objects that will be saved " "(vocab, transforms, embeddings, ...)",
    )  # "required" depending on build_vocab_only originally
    overwrite: bool = Field(default=False, description="Overwrite existing objects if any.")

    @field_validator("transforms_configs", mode="before")
    @classmethod
    def _str_to_dict(cls, v) -> Dict:
        if isinstance(v, str):
            if v == "None":
                return None  # not sure about this patch to allow forcing encoder/decoder to None (e.g. LM) # noqa: E501
            w = eval(v)
            return w
        return v

    def _get_all_transform(self):
        # might be good to rename at some point
        all_transforms = set(self.transforms)
        if self.data is not None:
            for cname, corpus in self.data.items():
                _transforms = set(corpus.transforms)
                if len(_transforms) != 0:
                    all_transforms.update(_transforms)
        if hasattr(self, "model") and getattr(getattr(self.model, "decoder", None), "lambda_align", 0.0) > 0.0:
            if not all_transforms.isdisjoint({"sentencepiece", "bpe", "onmt_tokenize"}):
                raise ValueError("lambda_align is not compatible with" " on-the-fly tokenization.")
            if not all_transforms.isdisjoint({"tokendrop", "prefix", "bart"}):
                raise ValueError("lambda_align is not compatible yet with" " potentiel token deletion/addition.")
        self._all_transform = all_transforms

    def _validate_vocab_config(self, build_vocab_only=False):
        # patch for n_sample default value around here?

        if build_vocab_only:
            if not self.share_vocab:
                assert self.tgt_vocab, "-tgt_vocab is required if not -share_vocab."
            return
        # validation when train:
        self.__class__._validate_file(self.src_vocab, info="src vocab")
        if not self.share_vocab:
            self.__class__._validate_file(self.tgt_vocab, info="tgt vocab")

        # Check embeddings stuff
        if self.both_embeddings is not None:
            assert (
                self.src_embeddings is None and self.tgt_embeddings is None
            ), "You don't need -src_embeddings or -tgt_embeddings \
                if -both_embeddings is set."

        if any(
            [
                self.both_embeddings is not None,
                self.src_embeddings is not None,
                self.tgt_embeddings is not None,
            ]
        ):
            assert self.embeddings_type is not None, "You need to specify an -embeddings_type!"
            assert (
                self.save_data
            ), "-save_data should be set if use \
                pretrained embeddings."

    @staticmethod
    def _validate_file(file_path, info):
        """Check `file_path` is valid or raise `IOError`."""
        if file_path == "dummy" or file_path.startswith("hf://"):
            # hack to allow creating objects with required fields
            pass
        elif not os.path.isfile(file_path):
            raise IOError(f"Please check path of your {info} file! ({file_path})")

    def _validate_data(self):
        """Parse corpora specified in data field of YAML file."""
        default_transforms = self.transforms
        if len(default_transforms) != 0 and len(self.data) > 0:
            # 0 datasets means we're loading a converted model
            logger.info(f"Default transforms (might be overridden downstream): {default_transforms}.")

        corpora = self.data

        for cname, corpus in corpora.items():
            # Check Transforms
            _transforms = corpus.transforms
            if _transforms is None:
                logger.info(f"Missing transforms field for {cname} data, " f"set to default: {default_transforms}.")
                corpus.transforms = default_transforms
            # Check path
            if corpus.path_src is None and corpus.path_txt is None:
                raise ValueError(
                    f"Corpus {cname} `path_src` or `path_txt` is required."
                    "Additionally, `path_tgt` is required for non language"
                    " modeling tasks."
                )
            elif corpus.path_src is not None:
                self.__class__._validate_file(corpus.path_src, info=f"{cname}/path_src")
                if corpus.path_tgt is None:
                    logger.debug("path_tgt is None, it should be set unless the task" " is language modeling")
                    corpora[cname] = corpus
                else:
                    self.__class__._validate_file(corpus.path_tgt, info=f"{cname}/path_tgt")
            if corpus.path_align is None:
                if hasattr(self, "model") and getattr(getattr(self.model, "decoder", None), "lambda_align", 0.0) > 0.0:
                    raise ValueError(f"Corpus {cname} alignment file path are " "required when lambda_align > 0.0")
                corpus.path_align = None
            else:
                self.__class__._validate_file(corpus.path_align, info=f"{cname}/path_align")
            # Check weight
            if corpus.weight is None:
                if cname != constants.CorpusName.VALID:
                    logger.warning(f"Corpus {cname}'s weight should be given." " We default it to 1 for you.")
                corpus.weight = 1
        if len(corpora) > 0:
            logger.info(f"Parsed {len(corpora)} corpora from -data.")
        # self.data = corpora
        self.__dict__["data"] = corpora  # skip validation to avoid recursion error

    # @model_validator(mode="after")
    # This needs to be called only after the config is fully up to date,
    # to prevent setting and keeping some unwanted values
    # (e.g. transforms [] for datasets)
    def _validate_data_config(self, build_vocab_only=False):
        if self.data is not None:  # patch to allow None data
            self._validate_data()
        self._get_all_transform()
        # this is manually triggered where needed, to allow instanciation of
        # TrainConfig without existing files (e.g. inference)
        # self._validate_vocab_config(build_vocab_only=build_vocab_only)
        return self

    @model_validator(mode="after")
    def _maybe_set_huggingface_model(self):
        if getattr(self, "model", None) is None:
            return self
        if self.model.huggingface_model is not None:
            if hasattr(self.transforms_configs, "huggingface_tokenize"):
                self.transforms_configs.huggingface_tokenize.huggingface_model = self.model.huggingface_model
        return self



import torch.nn as nn


class DecoderBase(nn.Module):
    """Abstract class for decoders.

    Args:
        attentional (bool): The decoder returns non-empty attention.
    """

    def __init__(self, attentional=True):
        super(DecoderBase, self).__init__()
        self.attentional = attentional
        # Decoder state
        self.state = {}

    @classmethod
    def from_config(cls, decoder_config, running_config=None, with_cross_attn=False):
        """Alternate constructor.

        Subclasses should override this method.
        """

        raise NotImplementedError

    def init_state(self, **kwargs):
        raise NotImplementedError

    def map_state(self, fn):
        raise NotImplementedError

    def detach_state(self):
        raise NotImplementedError



# some utilitary functions to facilitate cli tests
# and transition to config paradigm
import ast
import types
from typing import Union


def is_potentially_bool(field_type):
    if field_type is None:
        return False
    return bool in getattr(field_type, "__args__", [field_type])


def is_list(field_type):
    """This is very dirty and should not be kept in the future."""
    if field_type is None or isinstance(field_type, type(None)):
        return False
    # cases like "int | None" are types.UnionType instead of typing.Union
    elif isinstance(field_type, types.UnionType):
        return False
    elif getattr(field_type, "__origin__", None) == Union:
        return any([is_list(item) for item in field_type.__args__])
    else:
        return getattr(field_type, "_name", None) == "List"


def parse_type_annotation(annotation: str):
    try:
        # Parse the type annotation using ast
        tree = ast.parse(annotation, mode="eval")

        # Convert ast expression to a callable type
        return eval(compile(tree, filename="", mode="eval"))
    except (SyntaxError, TypeError):
        # Handle parsing errors
        return None


# test workaround to facilitate migration/testing
def add_model(parser, model):
    "Add Pydantic model to an ArgumentParser"
    fields = model.__fields__
    for name, field in fields.items():
        # special case for bool flags
        if is_potentially_bool(field.annotation):
            parser.add_argument(
                f"--{name}",
                f"-{name}",
                dest=name,
                action="store_true",
                default=field.default,
                help=field.description,
            )
        # we might also need one for nargs+
        elif is_list(field.annotation):
            parser.add_argument(
                f"--{name}",
                f"-{name}",
                dest=name,
                type=parse_type_annotation(field.annotation),
                nargs="+",
                default=field.default,
                help=field.description,
            )
        else:
            parser.add_argument(
                f"--{name}",
                f"-{name}",
                dest=name,
                type=parse_type_annotation(field.annotation),
                default=field.default,
                help=field.description,
                # help=field.field_info.description,
            )



from eole.bin import BaseBin
from eole.config.cli import add_model
from eole.config import get_non_default_values
from argparse import ArgumentParser
import os
import yaml


class RunBin(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--config",
            "-config",
            "-c",
            required=cls.require_config,
            help="Path of main YAML config file.",
        )
        add_model(parser, cls.config_class)

    @classmethod
    def build_config(cls, args):
        if args.config is not None:
            with open(args.config) as f:
                config_dict = yaml.safe_load(os.path.expandvars(f.read()))
        else:
            config_dict = {}

        # update config with non default values from command line args
        _parser = ArgumentParser()
        add_model(_parser, cls.config_class)
        defaults = vars(_parser.parse_args([]))
        stuff_to_update = get_non_default_values(args, defaults)
        config_dict.update(stuff_to_update)

        # pop extra fields added by argparse, not supported in pydantic configs
        if "bin" in config_dict.keys():
            config_dict.pop("bin")
        if "config" in config_dict.keys():
            config_dict.pop("config")

        config = cls.config_class(**config_dict)
        return config

    @classmethod
    def run(cls, args):
        raise NotImplementedError
